/* packet-wirebrowse.c
 *
 * Copyright 2011, Jakub Zawadzki <darkjames-ws@darkjames.pl>
 *
 * Google Summer of Code 2011 for The Honeynet Project
 * Mentors:
 *    Guillaume Arcas <guillaume.arcas (at) retiaire.org>
 *    Jeff Nathan <jeffnathan (at) gmail.com>
 *
 * $Id$
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <errno.h>
#include <glib.h>

#include <ctype.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <epan/dfilter/dfilter.h>
#include <color.h>
#include <color_filters.h>

#include <epan/epan_dissect.h>
#include <epan/epan.h>
#include <epan/packet.h>
#include <epan/prefs.h>
#include <epan/reassemble.h>
#include <epan/etypes.h>
#include <epan/expert.h>
#include <epan/report_err.h>
#include <epan/tap.h>

#include <wiretap/wtap.h>

static struct wtap *current_wth;

static emem_tree_t empty_tree;
static emem_tree_t *browse_frames_tree = &empty_tree;
static emem_tree_t *browse_conn_eth_tree = &empty_tree;
static emem_tree_t *browse_conn_tcp_tree = &empty_tree;
static emem_tree_t *browse_conn_udp_tree = &empty_tree;
static emem_tree_t *browse_conn_ip4_tree = &empty_tree;
static emem_tree_t *browse_conn_ip6_tree = &empty_tree;

static int
queue_frame(void *tapdata, packet_info *pinfo, epan_dissect_t *edt _U_, const void *data _U_)
{
	se_tree_insert32(*(emem_tree_t **) tapdata, pinfo->fd->num, pinfo->fd);
	return 0;
}

static void
tree_frames_retap(emem_tree_node_t *node)
{
	frame_data *fdata;
		
	if (!node)
		return;

	fdata = (frame_data *) node->data;

	if (node->left)
		tree_frames_retap(node->left);

	if (node->u.is_subtree == EMEM_TREE_NODE_IS_SUBTREE) {
		tree_frames_retap(node->data);
	} else {
		union wtap_pseudo_header pseudo_header;
		guint8 pd[WTAP_MAX_PACKET_SIZE+1];
		int err;
		gchar *err_info;

		if (wtap_seek_read(current_wth, fdata->file_off, &pseudo_header, pd, fdata->cap_len, &err, &err_info)) {
			epan_dissect_t edt;

			epan_dissect_init(&edt, FALSE, FALSE);
			tap_queue_init(&edt);
			epan_dissect_run(&edt, &pseudo_header, pd, fdata, NULL);
			tap_push_tapped_queue(&edt);
			epan_dissect_cleanup(&edt);
		} else
			g_free(err_info);
	}

	if (node->right)
		tree_frames_retap(node->right);
}

static void
browse_start(void)
{
	browse_frames_tree = se_tree_create(EMEM_TREE_TYPE_RED_BLACK, "Browse Frames Tree");
	browse_conn_eth_tree = se_tree_create(EMEM_TREE_TYPE_RED_BLACK, "Browse Eth Connection Tree");
	browse_conn_tcp_tree = se_tree_create(EMEM_TREE_TYPE_RED_BLACK, "Browse Tcp Connection Tree");
	browse_conn_udp_tree = se_tree_create(EMEM_TREE_TYPE_RED_BLACK, "Browse Udp Connection Tree");
	browse_conn_ip4_tree = se_tree_create(EMEM_TREE_TYPE_RED_BLACK, "Browse IPv4 Connection Tree");
	browse_conn_ip6_tree = se_tree_create(EMEM_TREE_TYPE_RED_BLACK, "Browse IPv6 Connection Tree");
}

static int
server_up(const char *ip, int port)
{
	/* IPv4 only */
	struct sockaddr_in sin;
	int fd;
	int err;
	int opt = 1;
	
	sin.sin_family = AF_INET;
	sin.sin_port = htons(port);
	sin.sin_addr.s_addr = inet_addr(ip);

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		return -1;

	/* not fatal */
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));

	if (bind(fd, (struct sockaddr *) &sin, sizeof(sin)))
		goto err;

	if (listen(fd, SOMAXCONN))
		goto err;
	
	return fd;
err:
	err = errno;
	close(fd);
	errno = err;
	return -1;
}

typedef struct {
	int fd;
	GString *input;
	GString *output;

} browse_client_t;

#define browse_putc(cli, c) \
	g_string_append_c(cli->output, c)

#define browse_puts(cli, buf) \
	g_string_append(cli->output, buf)

#define browse_printf(cli, format, args...) \
	g_string_append_printf(cli->output, format, args)

#define browse_write(cli, buf, count)	\
	g_string_append_len(cli->output, buf, count)

#define browse_putss(cli, buf) \
	browse_write(cli, buf, sizeof(buf)-1)

static void
browse_flush(browse_client_t *cli)
{
	if (cli->output->len > 1550) {
		GString *wr = cli->output;
		int wrote;

		do {
			wrote = write(cli->fd, wr->str, wr->len);

			if (wrote == -1) {
				if (errno == EINTR)
					continue;

				if (errno == EAGAIN)
					break;
			}
			if (wrote <= 0) {
				// XXX, handle it.
				return;
			}

			g_string_erase(wr, 0, wrote);

		} while (wr->len > 0);
	}
}

static void
browse_html_puts(browse_client_t *cli, const char *text)
{
	if (!text)
		return;
	while (*text) {
		switch (*text) {
			case '&':
				browse_putss(cli, "&amp;");
				break;
			case '<':
				browse_putss(cli, "&lt;");
				break;
			case '>':
				browse_putss(cli, "&gt;");
				break;
			case '"':
				browse_putss(cli, "&quot;");
				break;
			case '\'':
				browse_putss(cli, "&apos;");
				break;
			default:
				browse_putc(cli, *text);
		}
		text++;
	}
}

#define STATUS_200_OK        "HTTP/1.0 200 OK\r\n"
#define STATUS_404_NOT_FOUND "HTTP/1.0 404 Not Found\r\n"
#define CONTENT_HTML "Content-Type: text/html; charset=UTF-8\r\n"
#define CONTENT_XML  "Content-Type: text/xml\r\n"

static void
http_serve_common(browse_client_t *cli, const char *status, const char *content_type)
{
	static const char common[] = 
		"Server: WireBrowse\r\n"
		"Connection: keep-alive\r\n"
		"\r\n";

	browse_puts(cli, status);
	browse_puts(cli, content_type);
	browse_putss(cli, common);
}

static gboolean
tree_count_cb(void *node _U_, void *data)
{
	guint *count = (guint *) data;

	*count += 1;
	return FALSE;
}

static guint32
tree_count(emem_tree_t *emem_tree)
{
	guint32 count = 0;

	emem_tree_foreach(emem_tree, tree_count_cb, &count);
	return count;
}

static int
http_serve_index(browse_client_t *cli, const char *file _U_, const char *request _U_)
{
	http_serve_common(cli, STATUS_200_OK, CONTENT_HTML);

	browse_printf(cli, "<html><head><meta http-equiv='Content-Type' content='text/html; charset=UTF-8'><title>WireBrowse</title>\n"
				"<ul>"
					"<li><a href='frame'>Frame List (%u)</a></li>"
					"<li><a href='conn'>Connection List</a></li>"
				"</ul>",
					tree_count(browse_frames_tree)
				);
	return 0;
}

static void
http_serve_tree(browse_client_t *cli, proto_tree *tree, int level)
{
	static unsigned int ID = 0;

	proto_node *node;

	for (node = tree->first_child; node != NULL; node = node->next) {
		field_info *cur = PNODE_FINFO(node);
		guint32 color = 0x00ffffff;

		unsigned int id = ++ID;

		if (!cur)
			continue;

		/* colors from tree_cell_renderer */

        if (cur->hfinfo && cur->hfinfo->type == FT_PROTOCOL)
			color = 0xE5E5E5;	/* gray90 */

		if (FI_GET_FLAG(cur, PI_SEVERITY_MASK)) {
			switch (FI_GET_FLAG(cur, PI_SEVERITY_MASK)) {
				case (PI_CHAT):
					color = 0x80b7f7;	/* light blue */
					break;
				case (PI_NOTE):
					color = 0xa0ffff;	/* bright turquoise */
					break;
				case (PI_WARN):
					color = 0xf7f253;	/* yellow */
					break;
				case (PI_ERROR):
					color = 0xff5c5c;	/* pale red */
					break;
				default:
					g_assert_not_reached();
			}
		}

		browse_printf(cli, "<li onclick='hibytes(this, %u, %u)' style='background-color: #%.6x;'>", cur->start, cur->length, color);
		if (((proto_tree *) node)->first_child)
			browse_printf(cli, "<a href='javascript:xordisplay(\"u%u\")' style='margin-left: -15px;'>&#8658;</a>", id);
		if (cur->rep)
			browse_html_puts(cli, cur->rep->representation);
		else {
			gchar label_str[ITEM_LABEL_LENGTH];

			proto_item_fill_label(cur, label_str);
			browse_html_puts(cli, label_str);
		}
		browse_putss(cli, "</li>");

		if (((proto_tree *) node)->first_child) {
			browse_printf(cli, "<ul id='u%u' style='display: none;'>\n", id);
			http_serve_tree(cli, (proto_tree *) node, level+1);
			browse_putss(cli, "</ul>\n");
		}
	}
}

#if 0
static int
http_serve_file(browse_client_t *cli, const char *file, const char *request _U_)
{
	char buf[4096];
	FILE *f;
	int ret;

	if (!(f = fopen(file, "rb")))
		return 404;

	http_serve_common(cli, STATUS_200_OK, CONTENT_XML);

	while ((ret = fread(buf, 1, sizeof(buf), f)) > 0)
		browse_write(cli, buf, ret);

	fclose(f);
	return 0;
}
#endif

static guint32
color_to_hex(const color_t *color)
{
	guint32 c = 
		((guint32) (0xff0000 * (color->red / 65535.0))   & 0xff0000) | 
		((guint32) (0x00ff00 * (color->green / 65535.0)) & 0x00ff00) |
		((guint32) (0x0000ff * (color->blue / 65535.0))  & 0x0000ff);
	return c;
}

static void
http_serve_frame_list_cb(browse_client_t *cli, dfilter_t *filter, column_info *cinfo, emem_tree_node_t *node)
{
	frame_data *fdata;
		
	if (!node)
		return;

	fdata = (frame_data *) node->data;

	if (node->left)
		http_serve_frame_list_cb(cli, filter, cinfo, node->left);

	if (node->u.is_subtree == EMEM_TREE_NODE_IS_SUBTREE) {
		http_serve_frame_list_cb(cli, filter, cinfo, node->data);
	} else {
		union wtap_pseudo_header pseudo_header;
		guint8 pd[WTAP_MAX_PACKET_SIZE+1];
		int err;
		gchar *err_info;

		epan_dissect_t edt;
		int i;

		if (!wtap_seek_read(current_wth, fdata->file_off, &pseudo_header, pd, fdata->cap_len, &err, &err_info)) {
			g_free(err_info);
			return;
		}

		/* epan_dissect_init(&edt, TRUE, TRUE); */
		epan_dissect_init(&edt, filter != NULL, FALSE);
		col_custom_prime_edt(&edt, cinfo);
		if (filter)
			epan_dissect_prime_dfilter(&edt, filter);
		epan_dissect_run(&edt, &pseudo_header, pd, fdata, cinfo);

		if (!filter || dfilter_apply_edt(filter, &edt)) {
			if ((fdata->color_filter)||(fdata->flags.marked)||(fdata->flags.ignored)) {
				guint32 bgcolor;
				guint32 fgcolor;

				if (fdata->flags.ignored) {
					/* default; never here (packets not added to browse_frames_tree when ignored) */
					bgcolor = 0x00ffffff;
					fgcolor = 0x00808080;

				} else if (fdata->flags.marked) {
					/* default */
					fgcolor = 0x00ffffff;
					bgcolor = 0x00000000;

				} else {
					const color_filter_t *color_filter = fdata->color_filter;
					
					fgcolor = color_to_hex(&color_filter->fg_color);
					bgcolor = color_to_hex(&color_filter->bg_color);
				}

				browse_printf(cli, "<tr onclick='showframe(this, %u)' style='color: #%.6x;' bgcolor='#%.6x;'>", fdata->num, fgcolor, bgcolor);
			} else
				browse_printf(cli, "<tr onclick='showframe(this, %u)'>", fdata->num);

			epan_dissect_fill_in_columns(&edt, FALSE, TRUE);

			for (i = 0; i < cinfo->num_cols; i++) {
				browse_putss(cli, "<td>");
				switch (cinfo->col_fmt[i]) {
					case COL_NUMBER:
						browse_printf(cli, "<a onclick='return popframe(%u)' target='_blank' href='/frame/%u'>", fdata->num, fdata->num);
						browse_html_puts(cli, cinfo->col_data[i]);
						browse_putss(cli, "</a>");
						break;

					default:
						browse_html_puts(cli, cinfo->col_data[i]);
						break;
				}
				browse_putss(cli, "</td>");
			}
			browse_putss(cli, "</tr>\n");
			browse_flush(cli);
		}
		epan_dissect_cleanup(&edt);
	}

	if (node->right)
		http_serve_frame_list_cb(cli, filter, cinfo, node->right);
}

static void
http_serve_frame_list(browse_client_t *cli, dfilter_t *filter)
{
	struct timeval tv1, tv2;
	column_info cinfo;
	int i;

	build_column_format_array(&cinfo, 7, TRUE);

	browse_putss(cli, "<table width='100%'>");

	browse_putss(cli, "<thead><tr>");
	for (i = 0; i < cinfo.num_cols; i++) {
		browse_putss(cli, "<th>");
		browse_html_puts(cli, cinfo.col_title[i]);
		browse_putss(cli, "</th>");
	}
	browse_putss(cli, "</tr></thead>\n");

	browse_putss(cli, "<tbody style='overflow: auto; height: 400px;'>");
gettimeofday(&tv1, NULL);
	http_serve_frame_list_cb(cli, filter, &cinfo, browse_frames_tree->tree);
gettimeofday(&tv2, NULL);
g_print("List took: %lu ms\n", (tv2.tv_sec - tv1.tv_sec) * 1000 + (tv2.tv_usec - tv1.tv_usec)/1000);
	browse_putss(cli, "</tbody>");

	browse_putss(cli, "</table>");
}

static void
http_serve_frame_tree(browse_client_t *cli, frame_data *fdata, union wtap_pseudo_header *ph, const guint8 *pd)
{
	epan_dissect_t edt;

	epan_dissect_init(&edt, TRUE, TRUE);
	epan_dissect_run(&edt, ph, pd, fdata, NULL);

	browse_putss(cli, "<ul>");
	http_serve_tree(cli, edt.tree, 0);
	browse_putss(cli, "</ul>");

	epan_dissect_cleanup(&edt);
}

static void
http_serve_frame_bytes(browse_client_t *cli, const frame_data *fdata, const guint8 *pd)
{
	guint32 i = 0;
	guint use_digits;

	browse_putss(cli, "<pre>");
	if (fdata->cap_len >= 0x10000000)
		use_digits = 8; /* need all 8 digits */
	else if (fdata->cap_len >= 0x01000000)
		use_digits = 7; /* need 7 digits */
	else if (fdata->cap_len >= 0x00100000)
		use_digits = 6; /* need 6 digits */
	else if (fdata->cap_len >= 0x00010000)
		use_digits = 5; /* need 5 digits */
	else
		use_digits = 4; /* we'll supply 4 digits */

	while (i < fdata->cap_len) {
		static const gchar hexchars[16] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };
		guint j;
		guchar c;

		j = use_digits;
		do {
			j--;
			c = (i >> (j*4)) & 0xF;
			browse_putc(cli, hexchars[c]);
		} while (j != 0);

		/* Print the line number */
		browse_putss(cli, "  ");

		j   = i;
		/* Print the hex bit */
		while (i < j + 16) {
			if (i < fdata->cap_len) {
				browse_printf(cli, "<span id='b%u'>", i); /* serio nie da si inaczej?! */
				browse_putc(cli, hexchars[(pd[i] & 0xf0) >> 4]);
				browse_putc(cli, hexchars[pd[i] & 0x0f]);
				browse_putss(cli, "</span>");
			} else
				browse_putss(cli, "  ");
			i++;
			/* Inter byte space if not at end of line */
			if (i < j + 16) {
				browse_putc(cli, ' ');
				/* insert a space every BYTE_VIEW_SEP bytes */
				if( ( i % 8) == 0 )
					browse_putc(cli, ' ');
			}
		}

		/* Print some space at the end of the line */
		browse_putss(cli, "   ");

		/* Print the ASCII bit */
		i = j;
		while (i < j + 16) {
			if (i < fdata->cap_len) {
				c = pd[i];
				browse_printf(cli, "<span id='a%u'>", i); /* serio nie da si inaczej?! */
				browse_putc(cli, isprint(c) ? c : '.');
				browse_putss(cli, "</span>");
			} else {
				browse_putc(cli, ' ');
			}
			i++;
			if (i < j + 16) {
				/* insert a space every BYTE_VIEW_SEP bytes */
				if( ( i % 8) == 0 )
					browse_putc(cli, ' ');
			}
		}
		browse_putc(cli, '\n');
	}
	browse_putss(cli, "</pre>");
}

static int
http_serve_pbytes(browse_client_t *cli, const char *file _U_, const char *request)
{
	guint frame_no;
	frame_data *fdata;

	if (request[0] == '\0' || request[0] == '?')
		return 404;
	if (sscanf(request + 1, "%u", &frame_no) != 1)
		return 404;
	if (!(fdata = se_tree_lookup32(browse_frames_tree, frame_no)))
		return 404;
	{
		union wtap_pseudo_header pseudo_header;
		guint8      pd[WTAP_MAX_PACKET_SIZE+1];
		int err;
		gchar *err_info;

		if (!wtap_seek_read(current_wth, fdata->file_off, &pseudo_header, pd, fdata->cap_len, &err, &err_info)) {
			g_free(err_info);
			return -1;
		}

		http_serve_common(cli, STATUS_200_OK, CONTENT_HTML);
		http_serve_frame_bytes(cli, fdata, pd);
	}
	return 0;
}

static int
http_serve_flist(browse_client_t *cli, const char *file _U_, const char *request)
{
    dfilter_t *dfp;

	if (request[0] != '/')
		return 404;

	request++;

	g_print("%s\n", request);
    if (!dfilter_compile(request, &dfp)) {
		http_serve_common(cli, STATUS_200_OK, CONTENT_HTML);
		browse_putss(cli, "Invalid filter!!!");
		return 0;
	}

	http_serve_frame_list(cli, dfp);
	dfilter_free(dfp);
	return 0;
}

static int
http_serve_ptree(browse_client_t *cli, const char *file _U_, const char *request)
{
	guint frame_no;
	frame_data *fdata;

	if (request[0] == '\0' || request[0] == '?')
		return 404;
	if (sscanf(request + 1, "%u", &frame_no) != 1)
		return 404;
	if (!(fdata = se_tree_lookup32(browse_frames_tree, frame_no)))
		return 404;
	{
		union wtap_pseudo_header pseudo_header;
		guint8      pd[WTAP_MAX_PACKET_SIZE+1];
		int err;
		gchar *err_info;

		if (!wtap_seek_read(current_wth, fdata->file_off, &pseudo_header, pd, fdata->cap_len, &err, &err_info)) {
			g_free(err_info);
			return -1;
		}

		http_serve_common(cli, STATUS_200_OK, CONTENT_HTML);
		http_serve_frame_tree(cli, fdata, &pseudo_header, pd);
	}
	return 0;
}

static int
http_serve_filter(browse_client_t *cli, const char *file _U_, const char *request)
{
    dfilter_t   *dfp;

	if (request[0] != '/')
		return 404;

	http_serve_common(cli, STATUS_200_OK, CONTENT_HTML);
	request++;

	if (!request[0]) {
		/* empty */
		browse_putss(cli, "#ffffff");
		return 0;
	}

    if (dfilter_compile(request, &dfp)) {
		if (dfilter_deprecated_tokens(dfp))
			browse_putss(cli, "#ffffb0"); /* light yellow */
		else
			browse_putss(cli, "#b0ffb0"); /* light green */
		dfilter_free(dfp);
	} else
		browse_putss(cli, "#ffb0b0"); /* light red */

	return 0;
}

static int
http_serve_frame(browse_client_t *cli, const char *file _U_, const char *request)
{
	guint frame_no;
	frame_data *fdata;

	if (request[0] == '\0' || request[0] == '?') {
		http_serve_common(cli, STATUS_200_OK, CONTENT_HTML);

		browse_putss(cli, "<html><head><meta http-equiv='Content-Type' content='text/html; charset=UTF-8'><title>Frame List - WireBrowse</title>\n"
					"<style type='text/css'>\n"
						"ul { list-style-type: none; }\n"
						"thead { background-color:#eee; color:#666666; font-weight: bold; cursor: default; }\n"
						".sel { background-color: #0080FF !important; color: #000000 !important; }\n"
					"</style\n>"
					"<script type='text/javascript'>"
						"var lastpsel = null;\n"
						"var lastisel = null;\n"
						"function refilter(x) { var r = new XMLHttpRequest(); r.open('GET', '/flist/' + x, true); r.onreadystatechange = function () { if (r.readyState == 4) { document.getElementById('flist').innerHTML = r.responseText; } }; r.send(null); }\n"
						"function checkfilter(el) { var r = new XMLHttpRequest(); r.open('GET', '/checkfilter/' + el.value, true); r.onreadystatechange = function () { if (r.readyState == 4) { el.style.background = r.responseText; } }; r.send(null); }\n"
						"function getframe(u, x) { var r = new XMLHttpRequest(); r.open('GET', u + '/' + x, true); r.onreadystatechange = function () { if (r.readyState == 4) { document.getElementById(u).innerHTML = r.responseText; } }; r.send(null); }\n"
						"function showframe(el, x) { getframe('ptree', x); getframe('pbytes', x); if (lastpsel) lastpsel.className = ''; el.className = 'sel'; lastpsel = el; }\n"
						"function xordisplay(x) { var e = document.getElementById(x); if (e) { e.style.display = (e.style.display == \"block\") ? \"none\" : \"block\" } }\n"
						"function unhibytes2(pre) { var e; var c = 0; do { e = document.getElementById(pre + c); if (e) { e.style.backgroundColor = '#fff'; } c++; } while (e); }\n"
						"function hibytes2(pre, from, count) { unhibytes2(pre); while (count-- > 0) { var e = document.getElementById(pre + from); if (e) { e.style.backgroundColor = '#0080FF'; } from++; } }\n"
						"function hibytes(el, from, count) { hibytes2('a', from, count); hibytes2('b', from, count); if (lastisel) lastisel.className = ''; el.className = 'sel'; lastisel = el;  }\n"

						"function popframe(x) { var newwindow = window.open('frame/' + x, 'name', 'height=900,width=1000'); if (window.focus) newwindow.focus(); return false; }\n"
					"</script>\n");

		/* Toolz */
		browse_putss(cli, "<div>");
		browse_putss(cli, "Filter: <input type='text' onchange='refilter(this.value);' onkeyup='checkfilter(this)' style='width: 400px;' />");
		browse_putss(cli, "</div>");

		/* Packet List */
		browse_putss(cli, "<div id='flist' style='width: 100%; height: 450px; border: 2px solid red;'>");
		http_serve_frame_list(cli, NULL);
		browse_putss(cli, "</div>");

		/* Packet Details */
		browse_putss(cli, "<div id='ptree' style='overflow: auto; width: 100%; height: 300px; border: 2px solid red;'>");
		browse_putss(cli, "Frame not selected");
		browse_putss(cli, "</div>");

		/* Packet Bytes */
		browse_putss(cli, "<div id='pbytes' style='overflow: auto; width: 100%; height: 200px; border: 2px solid red;'><pre>");
		browse_putss(cli, "Frame not selected");
		browse_putss(cli, "</pre></div>");

		return 0;
	}
	if (sscanf(request + 1, "%u", &frame_no) != 1) {
		/* XXX */
		return 404;
	}

	if (!(fdata = se_tree_lookup32(browse_frames_tree, frame_no)))
		return 404;
	{
		union wtap_pseudo_header pseudo_header;
		guint8      pd[WTAP_MAX_PACKET_SIZE+1];
		int err;
		gchar *err_info;

		if (!wtap_seek_read(current_wth, fdata->file_off, &pseudo_header, pd, fdata->cap_len, &err, &err_info)) {
			g_free(err_info);
			return -1;
		}

		http_serve_common(cli, STATUS_200_OK, CONTENT_HTML);
		browse_printf(cli, "<html><head><meta http-equiv='Content-Type' content='text/html; charset=UTF-8'><title>Frame #%u - WireBrowse</title>\n"
					"<style type='text/css'>\n"
						"ul { list-style-type: none; }\n"
						".sel { background-color: #0080FF !important; color: #000000 !important; }\n"
					"</style\n>"
					"<script type='text/javascript'>"
						"var lastisel = null;\n"
						"function xordisplay(x) { var e = document.getElementById(x); if (e) { e.style.display = (e.style.display == \"block\") ? \"none\" : \"block\" } }\n"
						"function unhibytes2(pre) { var e; var c = 0; do { e = document.getElementById(pre + c); if (e) { e.style.backgroundColor = '#fff'; } c++; } while (e); }\n"
						"function hibytes2(pre, from, count) { unhibytes2(pre); while (count-- > 0) { var e = document.getElementById(pre + from); if (e) { e.style.backgroundColor = '#0080FF'; } from++; } }\n"
						"function hibytes(el, from, count) { hibytes2('a', from, count); hibytes2('b', from, count); if (lastisel) lastisel.className = ''; el.className = 'sel'; lastisel = el;  }\n"
					"</script>\n", frame_no);

		/* Packet Details */
		browse_putss(cli, "<div style='overflow: auto; width: 100%; height: 600px; border: 2px solid red;'>");
		http_serve_frame_tree(cli, fdata, &pseudo_header, pd);
		browse_putss(cli, "</div>");

		/* Packet Bytes */
		browse_putss(cli, "<div style='overflow: auto; width: 100%; height: 300px; border: 2px solid red;'>");
		http_serve_frame_bytes(cli, fdata, pd);
		browse_putss(cli, "</div>");
	}
	return 0;
}

typedef struct {
	address  src_address;    /**< source address */
	address  dst_address;    /**< destination address */
	guint32  src_port;       /**< source port */
	guint32  dst_port;       /**< destination port */
} conv_key_t;

typedef struct {
	conv_key_t key;
	guint32  port_type;      /**< port_type (e.g. PT_TCP) */

	guint32  rx_frames;      /**< number of received packets */
	guint32  tx_frames;      /**< number of transmitted packets */
	guint64  rx_bytes;       /**< number of received bytes */
	guint64  tx_bytes;       /**< number of transmitted bytes */

	nstime_t start_time;     /**< start time for the conversation */
	nstime_t stop_time;      /**< stop time for the conversation */
} conv_data_t;

static void
add_conversation_table_data(GHashTable *convtable, const packet_info *pinfo, const address *src, const address *dst, guint32 src_port, guint32 dst_port, int port_type_val)
{
	conv_data_t *conversation;
	conv_key_t existing_key;
	int cmp_ret;

	cmp_ret = CMP_ADDRESS(src, dst);
	if (cmp_ret == 0)
		cmp_ret = (src_port > dst_port) ? 1 : (src_port < dst_port) ? -1 : 0;

	if (cmp_ret > 0) {
		const address *tmp_a = src;
		guint32 tmp_p = src_port;

		src = dst;
		src_port = dst_port;
		dst = tmp_a;
		dst_port = tmp_p;
    }

	existing_key.src_address = *src;
	existing_key.dst_address = *dst;
	existing_key.src_port = src_port;
	existing_key.dst_port = dst_port;

	/* try to find it among the existing known conversations */
	if (!(conversation = g_hash_table_lookup(convtable, &existing_key))) {
		/* if we still dont know what conversation this is it has to be a new one
		   and we have to allocate it and append it to the end of the list */
		conversation = g_malloc(sizeof(conv_data_t));

		COPY_ADDRESS(&conversation->key.src_address, src);
		COPY_ADDRESS(&conversation->key.dst_address, dst);
		conversation->key.src_port = src_port;
		conversation->key.dst_port = dst_port;

		conversation->port_type = port_type_val;
		conversation->rx_frames = 0;
		conversation->tx_frames = 0;
		conversation->rx_bytes = 0;
		conversation->tx_bytes = 0;

		conversation->start_time = pinfo->fd->rel_ts;
		conversation->stop_time = pinfo->fd->rel_ts;

		g_hash_table_insert(convtable, &(conversation->key), conversation);
	}

	/* update the conversation struct */
	if (cmp_ret > 0) {
		conversation->rx_frames += 1;
		conversation->rx_bytes += pinfo->fd->pkt_len;
	} else {
		conversation->tx_frames += 1;
		conversation->tx_bytes += pinfo->fd->pkt_len;
	}

	if (nstime_cmp(&pinfo->fd->rel_ts, &conversation->stop_time) > 0)
		conversation->stop_time = pinfo->fd->rel_ts;
	else if (nstime_cmp(&pinfo->fd->rel_ts, &conversation->start_time) < 0)
		conversation->start_time = pinfo->fd->rel_ts;
}

#include <epan/dissectors/packet-eth.h>
static int
http_serve_conn_list_eth_cb(void *tapdata, packet_info *pinfo, epan_dissect_t *edt _U_, const void *data)
{
	const eth_hdr *ehdr = (const eth_hdr *) data;
	GHashTable *conv = (GHashTable *) tapdata;

	add_conversation_table_data(conv, pinfo, &ehdr->src, &ehdr->dst, 0, 0, PT_NONE);
	return 0;
}

#include <epan/dissectors/packet-ip.h>
static int
http_serve_conn_list_ip_cb(void *tapdata, packet_info *pinfo, epan_dissect_t *edt _U_, const void *data)
{
	const ws_ip *iph = (const ws_ip *) data;
	GHashTable *conv = (GHashTable *) tapdata;

	add_conversation_table_data(conv, pinfo, &iph->ip_src, &iph->ip_dst, 0, 0, PT_NONE);
	return 0;
}

#include <epan/dissectors/packet-ipv6.h>
static int
http_serve_conn_list_ip6_cb(void *tapdata, packet_info *pinfo, epan_dissect_t *edt _U_, const void *data)
{
	const struct ip6_hdr *ip6h = (const struct ip6_hdr *) data;
	GHashTable *conv = (GHashTable *) tapdata;
	address src;
	address dst;

	/* Addresses aren't implemented as 'address' type in struct ip6_hdr */
	src.type = dst.type = AT_IPv6;
	src.len  = dst.len = sizeof(struct e_in6_addr);
	src.data = &ip6h->ip6_src;
	dst.data = &ip6h->ip6_dst;

	add_conversation_table_data(conv, pinfo, &src, &dst, 0, 0, PT_NONE);
	return 0;
}

#include <epan/dissectors/packet-tcp.h>
static int
http_serve_conn_list_tcp_cb(void *tapdata, packet_info *pinfo, epan_dissect_t *edt _U_, const void *data)
{
	const struct tcpheader *tcphdr = (const struct tcpheader *) data;
	GHashTable *conv = (GHashTable *) tapdata;

	add_conversation_table_data(conv, pinfo, &tcphdr->ip_src, &tcphdr->ip_dst, tcphdr->th_sport, tcphdr->th_dport, PT_TCP);
	return 0;
}

#include <epan/dissectors/packet-udp.h>
static int
http_serve_conn_list_udp_cb(void *tapdata, packet_info *pinfo, epan_dissect_t *edt _U_, const void *data)
{
	const e_udphdr *udphdr = (const e_udphdr *) data;
	GHashTable *conv = (GHashTable *) tapdata;

	add_conversation_table_data(conv, pinfo, &udphdr->ip_src, &udphdr->ip_dst, udphdr->uh_sport, udphdr->uh_dport, PT_UDP);
	return 0;
}

static guint
conversation_hash(gconstpointer v)
{
	const conv_key_t *key = (const conv_key_t *)v;
	guint hash_val = 0;

	ADD_ADDRESS_TO_HASH(hash_val, &key->src_address);
	hash_val += key->src_port;
	ADD_ADDRESS_TO_HASH(hash_val, &key->dst_address);
	hash_val += key->dst_port;

	return hash_val;
}

/*
 * Compare two conversation keys for an exact match.
 */
static gboolean
conversation_match(gconstpointer v, gconstpointer w)
{
	const conv_key_t *v1 = (const conv_key_t *)v;
	const conv_key_t *v2 = (const conv_key_t *)w;

	return (v1->src_port == v2->src_port &&
			v1->dst_port == v2->dst_port &&
			ADDRESSES_EQUAL(&v1->src_address, &v2->src_address) &&
			ADDRESSES_EQUAL(&v1->dst_address, &v2->dst_address));
}

static void
conversation_free(gpointer data)
{
	conv_data_t *conv = data;

	g_free((gpointer)conv->key.src_address.data);
	g_free((gpointer)conv->key.dst_address.data);
	g_free(conv);
}

static const char *
ct_get_filter_addr(const address *addr)
{
	switch(addr->type){
		case AT_ETHER:
			return "eth.addr";
		case AT_IPv4:
			return "ip.addr";
		case AT_IPv6:
			return "ipv6.addr";
		default:
			return NULL;
	}
}

static const char *
ct_get_filter_port(int port_type_val)
{
	switch(port_type_val){
		case PT_TCP:
			return "tcp.port";
		case PT_UDP:
			return "udp.port";
		default:
			return NULL;
	}
	return NULL;
}

static void
http_serve_conv_list_cb(gpointer key _U_, gpointer value, gpointer user_data)
{
	const conv_data_t *conv = value;
	browse_client_t *cli = (browse_client_t *) user_data;
	char src_addr[MAX_ADDR_STR_LEN];
	char dst_addr[MAX_ADDR_STR_LEN];
	double duration_s;

	char *filter;

	address_to_str_buf(&conv->key.src_address, src_addr, MAX_ADDR_STR_LEN);
	address_to_str_buf(&conv->key.dst_address, dst_addr, MAX_ADDR_STR_LEN);

	if (conv->port_type != PT_NONE) {
		filter = 
			g_strdup_printf("%s==%s && %s == %u && %s==%s && %s == %u",
					ct_get_filter_addr(&conv->key.src_address),
					src_addr,
					ct_get_filter_port(conv->port_type),
					conv->key.src_port,
					ct_get_filter_addr(&conv->key.dst_address),
					dst_addr,
					ct_get_filter_port(conv->port_type),
					conv->key.dst_port);

	} else {
		filter = 
			g_strdup_printf("%s==%s && %s==%s",
					ct_get_filter_addr(&conv->key.src_address),
					src_addr,
					ct_get_filter_addr(&conv->key.dst_address),
					dst_addr);
	}
	/* XXX, escape filter */
	browse_printf(cli, "<tr onclick='selconn(this, \"%s\")'>", filter);
	g_free(filter);

	/* src */
	browse_putss(cli, "<td>");
	browse_html_puts(cli, src_addr);
	browse_putss(cli, "</td>");
	if (conv->port_type != PT_NONE)
		browse_printf(cli, "<td>%u</td>", conv->key.src_port);

	/* dest */
	browse_putss(cli, "<td>");
	browse_html_puts(cli, dst_addr);
	browse_putss(cli, "</td>");
	if (conv->port_type != PT_NONE)
		browse_printf(cli, "<td>%u</td>", conv->key.dst_port);

	browse_printf(cli, "<td>%u</td>", conv->tx_frames + conv->rx_frames);
	browse_printf(cli, "<td>%" G_GUINT64_FORMAT "</td>", conv->tx_bytes + conv->rx_bytes);
	browse_printf(cli, "<td>%u</td>", conv->tx_frames);
	browse_printf(cli, "<td>%" G_GUINT64_FORMAT "</td>", conv->tx_bytes);
	browse_printf(cli, "<td>%u</td>", conv->rx_frames);
	browse_printf(cli, "<td>%" G_GUINT64_FORMAT "</td>", conv->rx_bytes);

	duration_s = nstime_to_sec(&conv->stop_time) - nstime_to_sec(&conv->start_time);

	browse_printf(cli, "<td>%s</td>", rel_time_to_secs_str(&conv->start_time));
	browse_printf(cli, "<td>%.4f</td>", duration_s);
	browse_printf(cli, "<td>%.2f</td>", (gint64) conv->tx_bytes * 8 / duration_s);
	browse_printf(cli, "<td>%.2f</td>",  (gint64) conv->rx_bytes * 8 / duration_s);
	browse_putss(cli, "</tr>\n");
	browse_flush(cli);
}

static void
http_serve_conv_list(browse_client_t *cli, const char *tap, const tap_packet_cb tapfunc, emem_tree_t *tree, gboolean show_ports)
{
	browse_putss(cli, "<table width='100%'>");

	browse_printf(cli, "<thead><tr>"
						"<th>Address A</th>"
						"%s"
						"<th>Address B</th>"
						"%s"
						"<th>Packets</th>"
						"<th>Bytes</th>"
						"<th>Packets A &#x2192; B</th>"
						"<th>Bytes A &#x2192; B</th>"
						"<th>Packets A &#x2190; B</th>"
						"<th>Bytes A &#x2190; B</th>"
						"<th>Rel Start</th>"
						"<th>Duration</th>"
						"<th>bps A &#x2192; B</th>"
						"<th>bps A &#x2190; B</th>"
					"</tr></thead>\n",
					(show_ports) ? "<th>Port A</th>" : "",
					(show_ports) ? "<th>Port B</th>" : "");

	browse_putss(cli, "<tbody style='overflow: auto; height: 450px;'>");
	{
		GHashTable *hashtable = g_hash_table_new_full(conversation_hash, conversation_match, NULL, conversation_free);
		struct timeval tv1, tv2;

		register_tap_listener(tap, hashtable, NULL, TL_IS_DISSECTOR_HELPER, NULL, tapfunc, NULL);
	gettimeofday(&tv1, NULL);
		tree_frames_retap(tree->tree);
	gettimeofday(&tv2, NULL);
	g_print("Retap took: %lu ms\n", (tv2.tv_sec - tv1.tv_sec) * 1000 + (tv2.tv_usec - tv1.tv_usec)/1000);
		remove_tap_listener(hashtable);

		g_hash_table_foreach(hashtable, http_serve_conv_list_cb, cli);

		g_hash_table_destroy(hashtable);
	}
	browse_putss(cli, "</tbody>");

	browse_putss(cli, "</table>");
}

static int
http_serve_conn(browse_client_t *cli, const char *file _U_, const char *request)
{
	tap_packet_cb packet_func;
	gboolean show_ports = FALSE;
	emem_tree_t *tree;

	if (request[0] == '\0')
		request = "/tcp";

	if (request[0] != '/')
		return 404;

	request++;

	if (!strcmp(request, "tcp")) {
		tree = browse_conn_tcp_tree;
		packet_func = http_serve_conn_list_tcp_cb;
		show_ports = TRUE;
	} else if (!strcmp(request, "udp")) {
		tree = browse_conn_udp_tree;
		packet_func = http_serve_conn_list_udp_cb;
		show_ports = TRUE;
	} else if (!strcmp(request, "eth")) {
		tree = browse_conn_eth_tree;
		packet_func = http_serve_conn_list_eth_cb;
	} else if (!strcmp(request, "ip")) {
		tree = browse_conn_ip4_tree;
		packet_func = http_serve_conn_list_ip_cb;
	} else if (!strcmp(request, "ipv6")) {
		tree = browse_conn_ip6_tree;
		packet_func = http_serve_conn_list_ip6_cb;
	} else
		return 404;
		
	http_serve_common(cli, STATUS_200_OK, CONTENT_HTML);

	browse_putss(cli, "<html><head><meta http-equiv='Content-Type' content='text/html; charset=UTF-8'><title>Connection List - WireBrowse</title>\n"
				"<style type='text/css'>\n"
					"ul { list-style-type: none; }\n"

					"#foo li { display: inline; margin: 0 2px 0 0; }\n"
					"#foo a { padding: 0 1em; text-decoration: none; background-color: #ccc }\n"
					"#foo a:hover { background-color: #00e0FF; }\n"

					"thead { background-color:#eee; color:#666666; font-weight: bold; cursor: default; }\n"
					".sel { background-color: #0080FF !important; color: #000000 !important; }\n"
				"</style>\n"
				"<script type='text/javascript'>"
					"var lastisel = null;\n"
					"function refilter(x) { var r = new XMLHttpRequest(); r.open('GET', '/flist/' + x, true); r.onreadystatechange = function () { if (r.readyState == 4) { document.getElementById('flist').innerHTML = r.responseText; } }; r.send(null); }\n"
					"function selconn(el, filter) { refilter(filter); if (lastisel) lastisel.className = ''; el.className = 'sel'; lastisel = el; }\n"

					"function showframe(el, x) { }\n"
					"function popframe(x) { return true; }"
				"</script>\n");

	/* Tabz */
	browse_printf(cli, "<div id='foo'>"
						"<ul>"
							"<li><a %shref='/conn/eth'>Ethernet: %u</a></li>"
							"<li><a %shref='/conn/ip'>IPv4: %u</a></li>"
							"<li><a %shref='/conn/ipv6'>IPv6: %u</a></li>"
							"<li><a %shref='/conn/tcp'>TCP: %u</a></li>"
							"<li><a %shref='/conn/udp'>UDP: %u</a></li>"
						"</ul>"
					"</div>",
						(tree == browse_conn_eth_tree) ? "class='sel' " : "", tree_count(browse_conn_eth_tree),
						(tree == browse_conn_ip4_tree) ? "class='sel' " : "", tree_count(browse_conn_ip4_tree),
						(tree == browse_conn_ip6_tree) ? "class='sel' " : "", tree_count(browse_conn_ip6_tree),
						(tree == browse_conn_tcp_tree) ? "class='sel' " : "", tree_count(browse_conn_tcp_tree),
						(tree == browse_conn_udp_tree) ? "class='sel' " : "", tree_count(browse_conn_udp_tree)
					);

	/* List */
	browse_putss(cli, "<div style='overflow: auto; width: 100%; height: 500px; border: 2px solid red;'>");
	http_serve_conv_list(cli, request, packet_func, tree, show_ports);
	browse_putss(cli, "</div>");

	browse_putss(cli, "<div id='flist' style='overflow: auto; width: 100%; height: 450px; border: 2px solid red;'>");
	browse_putss(cli, "Select connection to get list list of packets\n");
	browse_putss(cli, "</div>");
	return 0;
}

static void
wirebrowse_handle_http(browse_client_t *cli, char *data)
{
	static const struct {
		const char *base;
		gboolean isdir;
		int (*serve)(browse_client_t *cli, const char *base, const char *req);

	} pages[] = {
		{ "/frame", TRUE, http_serve_frame },

		{ "/checkfilter", TRUE, http_serve_filter },
		{ "/flist", TRUE, http_serve_flist },
		{ "/pbytes", TRUE, http_serve_pbytes },
		{ "/ptree", TRUE, http_serve_ptree },

		{ "/conn", TRUE, http_serve_conn },

		{ "/", FALSE, http_serve_index },

//		{ "http://", FALSE, http_serve_proxy }
	};

	static const char worka[] = 
		"Not found!";

	char *request = data;	/* first line */
	char *request_end = strstr(request, "\r\n");

	*request_end = '\0';

	if (g_str_has_prefix(request, "GET ")) {
		char *url;
		char *protocol = NULL;
		unsigned int i;

		int code = -1;

		url = request + 4;

		if ((protocol = strchr(url, ' '))) {
			*protocol = '\0';
			protocol+=1;
		}

		for (i = 0; i < array_length(pages); i++) {
			size_t baselen = strlen(pages[i].base);

			if (!strncmp(url, pages[i].base, baselen) && (url[baselen] == '\0' || (pages[i].isdir && url[baselen] == '/') || url[baselen] == '?')) {
				char *p = url + baselen;

				while (*p) {
					char *k = p++;

					if (*k == '%') {
						int res = 0;

						if (p[0] && p[1] && g_ascii_xdigit_value(p[0]) != -1 && g_ascii_xdigit_value(p[1]) != -1) {
							res = g_ascii_xdigit_value(*p++) << 4;
							res |= g_ascii_xdigit_value(*p++);
						}

						k[0] = res;
						g_memmove(k + 1, p, strlen(p)+1);
					}
				}
				code = pages[i].serve(cli, pages[i].base + 1, url + baselen);
				break;
			}
		}

		if (code == 404 || code == -1) {
			http_serve_common(cli, STATUS_404_NOT_FOUND, CONTENT_HTML);
			g_print("%s %s\n", url, protocol);
			browse_putss(cli, worka);
		}

	} else {
		g_print("%s\n", request);

		http_serve_common(cli, STATUS_404_NOT_FOUND, CONTENT_HTML); // XXX
		browse_putss(cli, worka);
	}
}

static gboolean
wirebrowse_write_socket(GIOChannel *source, GIOCondition condition, gpointer data)
{
	int fd = g_io_channel_unix_get_fd(source);

	GString *wr = (GString *) data;
	int wrote;

	do {
		wrote = write(fd, wr->str, wr->len);

		if (wrote == -1) {
			if (errno == EINTR)
				continue;

			if (errno == EAGAIN)
				break;
		}
		if (wrote <= 0)
			return FALSE;

		g_string_erase(wr, 0, wrote);

	} while (wr->len > 0);

	return (wr->len != 0);
}

static gboolean
wirebrowse_read_socket(GIOChannel *io, GIOCondition condition, gpointer data)
{
	browse_client_t *cli = (browse_client_t *) data;

	int fd = g_io_channel_unix_get_fd(io);
	char buf[1024];

	if (condition & G_IO_IN) {
		int ret;

		while ((ret = read(fd, buf, sizeof(buf))) > 0)
			g_string_append_len(cli->input, buf, ret);

		if (ret == -1 && errno != EINTR && errno != EAGAIN)
			return FALSE;
	}

	if ((strstr(cli->input->str, "\r\n\r\n"))) {
		wirebrowse_handle_http(cli, cli->input->str);

		g_io_add_watch(io, G_IO_OUT, wirebrowse_write_socket, cli->output);

		cli->output = NULL;
		return FALSE;
	}

	return TRUE;
}

static gboolean
wirebrowse_accept_socket(GIOChannel *source, GIOCondition condition, gpointer data _U_)
{
	int sfd = g_io_channel_unix_get_fd(source);

	struct sockaddr_in sin;
	socklen_t sinlen;
	int fd;
	
	if (condition & G_IO_IN) {
		GIOChannel *io;
		fd = accept(sfd, (struct sockaddr *) &sin, &sinlen);

		g_print("wirebrowse_accept_socket(): %d\n", fd);

		if (fd != -1) {
			browse_client_t *client = g_malloc(sizeof(browse_client_t));

			client->fd = fd;
			client->input = g_string_new(NULL);
			client->output = g_string_new(NULL);

			io = g_io_channel_unix_new(fd);
			g_io_channel_set_flags(io, G_IO_FLAG_NONBLOCK, NULL);
			g_io_channel_set_close_on_unref(io, TRUE);
			g_io_add_watch(io, G_IO_IN, wirebrowse_read_socket, client);
			g_io_channel_unref(io);
		}
	}
	return TRUE;
}

extern
void
proto_reg_handoff_wirebrowse(void)
{
	GIOChannel *io;
	int fd;
	
	fd = server_up("127.0.0.1", 8080);

	g_print("proto_reg_handoff_wirebrowse(): %d\n", fd);

	if (fd == -1) {
		report_failure("proto_reg_handoff_wirebrowse() server up fail: %s\n", g_strerror(errno));
		return;
	}

	io = g_io_channel_unix_new(fd);
	g_io_channel_set_close_on_unref(io, TRUE);
	g_io_add_watch(io, G_IO_IN, wirebrowse_accept_socket, NULL);
	g_io_channel_unref(io);

	register_tap_listener("frame", &browse_frames_tree, NULL, TL_IS_DISSECTOR_HELPER, NULL, queue_frame, NULL);
	register_tap_listener("eth", &browse_conn_eth_tree, NULL, TL_IS_DISSECTOR_HELPER, NULL, queue_frame, NULL);
	register_tap_listener("tcp", &browse_conn_tcp_tree, NULL, TL_IS_DISSECTOR_HELPER, NULL, queue_frame, NULL);
	register_tap_listener("udp", &browse_conn_udp_tree, NULL, TL_IS_DISSECTOR_HELPER, NULL, queue_frame, NULL);
	register_tap_listener("ip", &browse_conn_ip4_tree, NULL, TL_IS_DISSECTOR_HELPER, NULL, queue_frame, NULL);
	register_tap_listener("ipv6", &browse_conn_ip6_tree, NULL, TL_IS_DISSECTOR_HELPER, NULL, queue_frame, NULL);
}

extern
void
proto_register_wirebrowse(void)
{
	g_print("proto_register_wirebrowse()\n");
	register_init_routine(browse_start);
}

/* Hack,
 *    we're registering wtap open routine to get wtap handle ;-)
 */
static gint
wirebrowse_wtap_open(struct wtap *wtap, int *err _U_, char **err_info _U_)
{
	g_print("wirebrowse_wtap_open(): %p\n", wtap);
	current_wth = wtap;
	return 0;
}

extern
void 
wtap_register_wirebrowse_helper(void) {
	wtap_register_open_routine(wirebrowse_wtap_open, TRUE);
}
