/* packet-wireshav.c
 *
 * Copyright 2011, Jakub Zawadzki <darkjames-ws@darkjames.pl>
 *
 * Google Summer of Code 2011 for The Honeynet Project
 * Mentors:
 *    Guillaume Arcas <guillaume.arcas (at) retiaire.org>
 *    Jeff Nathan <jeffnathan (at) gmail.com>
 *
 * $Id$
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Antiviruses known to work:
 *
 * XXX, here list all windows/linux/unix antiviruses which can work with this plugin ;-)
 *      For now we support only clamav.
 *
 * (lib)clamav:
 *   app-antivirus/clamav-0.97
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define HAVE_LIBCLAMAV

#include <stdio.h>
#include <errno.h>
#include <glib.h>
#include <epan/packet.h>
#include <epan/prefs.h>
#include <epan/reassemble.h>
#include <epan/etypes.h>
#include <epan/expert.h>
#include <epan/report_err.h>
#include <epan/tap.h>
#include <epan/tfs.h>

#ifdef HAVE_LIBCLAMAV
# include <sys/types.h>
# include <sys/socket.h>
# include <sys/un.h>

# include <clamav.h>

# define CLAMPATH "unix:/var/run/clamav/clamd.sock"
#endif

/* tfs_detected_not_detected export fixed in r37814 */
static const true_false_string _tfs_detected_not_detected = { "Detected", "Not Detected" };

typedef struct {
	char *uri;
	char *type;
	char *sha256;

#ifdef HAVE_LIBCLAMAV
	char *clamav;
#endif

} av_scan_t;


static int proto_av = -1;

static int hf_av_uri = -1;
static int hf_av_sha = -1;
static int hf_av_virus = -1;
static int hf_av_clamav = -1;

static int ett_av = -1;

static emem_tree_t *av_files_tree;

static void 
av_dissector(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree) 
{
	av_scan_t *scanv;

	if ((scanv = se_tree_lookup32(av_files_tree, pinfo->fd->num))) {
		proto_tree *av_tree = NULL;
		proto_item *ti = NULL;
		gboolean found_virus = FALSE;

#ifdef HAVE_LIBCLAMAV
		found_virus |= (scanv->clamav != NULL);
#endif

		if (tree) {
			ti = proto_tree_add_protocol_format(tree, proto_av, tvb, 0, -1, "AV: %s (%s)", scanv->uri, scanv->type);
			av_tree = proto_item_add_subtree(ti, ett_av);
		}

		proto_tree_add_string(av_tree, hf_av_uri, tvb, 0, 0, scanv->uri);
		proto_tree_add_string(av_tree, hf_av_sha, tvb, 0, 0, scanv->sha256);
		proto_tree_add_boolean(av_tree, hf_av_virus, tvb, 0, 0, found_virus);

#ifdef HAVE_LIBCLAMAV
		if (scanv->clamav) 
			proto_tree_add_string(av_tree, hf_av_clamav, tvb, 0, 0, scanv->clamav);

		proto_item_append_text(ti, " [ClamAV: %s]", (scanv->clamav) ? scanv->clamav : "Clean");
#endif
	}
}

#ifdef HAVE_LIBCLAMAV

static struct cl_engine *clamav_engine;
static gboolean clamav_use_socket = FALSE;

static int
clamav_socket_init(const char *path)
{
	if (g_str_has_prefix(path, "unix:")) {
		struct sockaddr_un sun;
		int fd;
		
		sun.sun_family = AF_UNIX;
		if (g_strlcpy(sun.sun_path, path + 5, sizeof(sun.sun_path)) >= sizeof(sun.sun_path))
			return -1;

		if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
			return -1;

		if (connect(fd, (struct sockaddr *) &sun, sizeof(sun))) {
			int err = errno;

			close(fd);
			errno = err;
			return -1;
		}
		return fd;
	}

	if (g_str_has_prefix(path, "tcp:")) {
		/* XXX */
	}
	/* errno = ENOTSUP; */
	return -1;
}

static struct cl_engine *
clamav_engine_init(void)
{
	struct cl_engine *engine;
	unsigned int sigs = 0;
	int ret;

	if ((ret = cl_init(CL_INIT_DEFAULT)) != CL_SUCCESS) {
		printf("Can't initialize libclamav: %s\n", cl_strerror(ret));
		return NULL;
	}

	if (!(engine = cl_engine_new())) {
		printf("Can't create new engine\n");
		return NULL;
	}

	/* load all available databases from default directory */
	if ((ret = cl_load(cl_retdbdir(), engine, &sigs, CL_DB_STDOPT)) != CL_SUCCESS) {
		printf("cl_load: %s\n", cl_strerror(ret));
		cl_engine_free(engine);
		return NULL;
	}

	printf("Loaded %u signatures.\n", sigs);

	/* build engine */
	if ((ret = cl_engine_compile(engine)) != CL_SUCCESS) {
		printf("Database initialization error: %s\n", cl_strerror(ret));;
		cl_engine_free(engine);
		return NULL;
	}

	return engine;
}

static gboolean
full_write(int fd, const void *data, guint32 len)
{
	int res;

	/* split to 4 K */
	while (len > 4096U) {
		if ((res = write(fd, data, 4096)) != 4096)
			return FALSE;

		len -= 4096U;
	}

	if (len > 0) {
		if ((res = write(fd, data, len)) != (int) len)
			return FALSE;
	}

	return TRUE;
}

static const char *
clamav_scan_socket_stream(const char *data, guint32 len)
{
	static char rbuf[128];
	int fd = clamav_socket_init(CLAMPATH);
	guint32 len_tmp;

	if (fd == -1)
		return "cant-connect!";

	if (!full_write(fd, "zINSTREAM", 10))
		goto err;

	len_tmp = g_htonl(len);
	if (!full_write(fd, &len_tmp, sizeof(len_tmp)))
		goto err;

	if (!full_write(fd, data, len))
		goto err;

	len_tmp = 0;
	if (!full_write(fd, &len_tmp, sizeof(len_tmp)))
		goto err;

#if 0
	if(ret == CL_VIRUS) {
		if(context.virsize)
			detstats_add(virname, "NOFNAME", context.virsize, context.virhash);
		if(context.virsize && optget(opts, "ExtendedDetectionInfo")->enabled)
			mdprintf(odesc, "stream: %s(%s:%llu) FOUND%c", virname, context.virhash, context.virsize, term);
		else
			mdprintf(odesc, "stream: %s FOUND%c", virname, term);
		virusaction("stream", virname, opts);
	} else if(ret != CL_CLEAN) {
		if(retval == 1)
			mdprintf(odesc, "stream: %s ERROR%c", cl_strerror(ret), term);
	}
#endif

	len_tmp = read(fd, rbuf, sizeof(rbuf)-1);
	if (len_tmp > 0) {
		const char *buf = rbuf;

		rbuf[len_tmp] = '\0';
		close(fd);

		if (g_str_has_prefix(buf, "stream: ")) {
			buf = buf + sizeof("stream: ") - 1;

			if (!strcmp(buf, "OK"))
				return NULL;
		}
		if (buf[0]) {
			g_print("%s\n", buf);
			return buf;
		}
		return NULL;
	}
err:
	close(fd);
	return "err-write";
}

static const char *
clamav_scan_engine(const struct cl_engine *engine, const char *path)
{
	const char *virname;
	unsigned long int size = 0;
	int ret;

	ret = cl_scanfile(path, &virname, &size, engine, CL_SCAN_STDOPT);

	/* or:
	   int fd = open(path, O_RDONLY);
	   if (fd == -1)
	     return "can't open";
	   ret = cl_scandesc(fd, &virname, &size, engine, CL_SCAN_STDOPT); 
	   close(fd); 
	 */
	switch (ret) {
		case CL_CLEAN:
			return NULL;

		case CL_VIRUS:
			return virname;

		/* return error */
		default:
			return cl_strerror(ret);
	}
}
#endif

static void
scan_common(const packet_info *pinfo, const char *uri, const char *content_type, const guint8 *data, guint32 len)
{
	int fd;
	char *tmppath;
	const char *res;
	guint32 org_len = len;
	GChecksum *sha256_checksum;

	av_scan_t *scan;

	g_print("%d: %s %s %u\n", pinfo->fd->num, uri, content_type ? content_type : "(null)", len);

	// content_type image/jpeg, etc...
	if (content_type && strstr("tutaj_pref", content_type))
		return;

	if ((scan = se_tree_lookup32(av_files_tree, pinfo->fd->num)))
		return;

	fd = g_file_open_tmp("wireshav-XXXXXX", &tmppath, NULL);
	if (fd == -1)
		return;

	while (len > 0) {
		int tmplen = MIN(len, 4096);
		int wrote;

		if ((wrote = write(fd, data, tmplen)) != tmplen)
			goto err;
		len -= tmplen;
	}

	scan = se_alloc0(sizeof(av_scan_t));
	scan->uri = se_strdup(uri);
	scan->type = se_strdup(content_type);

#ifdef HAVE_LIBCLAMAV
	if ((res = (clamav_use_socket) ? 
			clamav_scan_socket_stream(data, org_len) :
			clamav_scan_engine(clamav_engine, tmppath)))
	{
		scan->clamav = se_strdup(res);
	}
#endif

	if ((sha256_checksum = g_checksum_new(G_CHECKSUM_SHA256))) {
		g_checksum_update(sha256_checksum, data, org_len);
		scan->sha256 = se_strdup(g_checksum_get_string(sha256_checksum));
		g_checksum_free(sha256_checksum);
	}

	se_tree_insert32(av_files_tree, pinfo->fd->num, scan);
err:
	close(fd);
	g_unlink(tmppath);

	return;
}

#include <epan/dissectors/packet-http.h>
static int
scan_http_packet(void *tapdata _U_, packet_info *pinfo, epan_dissect_t *edt _U_, const void *data)
{
	const http_eo_t *eo_info = (const http_eo_t *) data;
	char uri[1024];

	if (eo_info) {
		g_snprintf(uri, sizeof(uri), "http://%s%s", eo_info->hostname, eo_info->filename);
		scan_common(pinfo, uri, eo_info->content_type, eo_info->payload_data, eo_info->payload_len);
	}
	return 0;
}

#include <epan/dissectors/packet-imf.h>
static int
scan_imf_packet(void *tapdata _U_, packet_info *pinfo, epan_dissect_t *edt _U_, const void *data)
{
	const imf_eo_t *eo_info = (const imf_eo_t *) data;
	char uri[1024];

	if (eo_info) {
		g_snprintf(uri, sizeof(uri), "Message-ID:%s", eo_info->msg_id);
		scan_common(pinfo, uri, "message/rfc822", eo_info->payload_data, eo_info->payload_len);
	}
	return 0;
}

#include <epan/dissectors/packet-smb.h>
static int
scan_smb_packet(void *tapdata _U_, packet_info *pinfo, epan_dissect_t *edt _U_, const void *data)
{
	const smb_eo_t *eo_info = (const smb_eo_t *) data;
	char uri[1024];

	if (eo_info) {
/*
        guint8 cmd;
        int     tid,uid,fid;
        guint32  pkt_num;
        int fid_type;
        gint64  end_of_file;
        gchar   *content_type;
        guint64 smb_file_offset;
        guint32 smb_chunk_len;
 */
		g_snprintf(uri, sizeof(uri), "%s\\%s", eo_info->hostname, eo_info->filename);
		scan_common(pinfo, uri, NULL, eo_info->payload_data, eo_info->payload_len);
	}
	return 0;
}

static void
av_start(void)
{
	av_files_tree = se_tree_create(EMEM_TREE_TYPE_RED_BLACK, "AV Scanned Files Tree");
}

extern
void
proto_reg_handoff_wireshav(void)
{
#ifdef HAVE_LIBCLAMAV
	int clamav_fd = clamav_socket_init(CLAMPATH);
	if (clamav_fd != -1) {
		clamav_use_socket = TRUE;
		close(clamav_fd);
	} else {
		clamav_use_socket = FALSE;
		clamav_engine = clamav_engine_init();
	}
#endif

	register_tap_listener("http_eo", NULL, NULL, TL_IS_DISSECTOR_HELPER, NULL, scan_http_packet, NULL);
	register_tap_listener("imf_eo", NULL, NULL, TL_IS_DISSECTOR_HELPER, NULL, scan_imf_packet, NULL);
	register_tap_listener("smb_eo", NULL, NULL, TL_IS_DISSECTOR_HELPER, NULL, scan_smb_packet, NULL);
}

extern
void
proto_register_wireshav(void)
{
	static hf_register_info hf[] = {
		{ &hf_av_uri,
			{ "URI", "av.uri", FT_STRINGZ, BASE_NONE, NULL, 0x00,
			NULL, HFILL } },

		{ &hf_av_sha,
			{ "SHA-256", "av.sha", FT_STRINGZ, BASE_NONE, NULL, 0x00,
			NULL, HFILL } },

		{ &hf_av_virus,
			{ "Virus", "av.virus", FT_BOOLEAN, BASE_NONE, TFS(&_tfs_detected_not_detected), 0x00,
			NULL, HFILL } },

		{ &hf_av_clamav,
			{ "Clamav Antivirus", "av.clamav", FT_STRINGZ, BASE_NONE, NULL, 0x00,
			NULL, HFILL } }
	};

	static gint *ett[] = {
		&ett_av
	};

	dissector_handle_t av_handle;

	proto_av = proto_register_protocol("Wireshark AV Plugin", "WireshAv", "av");
	g_print("proto_register_wireavork()\n");
	proto_register_field_array(proto_av, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	av_handle = create_dissector_handle(av_dissector, proto_av);

	register_init_routine(av_start);
	register_postdissector(av_handle);
}

