/* packet-wireshnork.c
 *
 * Copyright 2011, Jakub Zawadzki <darkjames-ws@darkjames.pl>
 *
 * Google Summer of Code 2011 for The Honeynet Project
 * Mentors:
 *    Guillaume Arcas <guillaume.arcas (at) retiaire.org>
 *    Jeff Nathan <jeffnathan (at) gmail.com>
 *
 * $Id$
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Snort version known to run:
 *
 * net-analyzer/snort-2.9.0.5  USE="decoder-preprocessor-rules dynamicplugin ipv6 threads zlib"
 * net-analyzer/snort-2.8.6  USE="decoder-preprocessor-rules dynamicplugin ipv6 zlib"
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <errno.h>
#include <glib.h>
#include <epan/packet.h>
#include <epan/prefs.h>
#include <epan/reassemble.h>
#include <epan/etypes.h>
#include <epan/expert.h>
#include <epan/report_err.h>

#include <wiretap/wtap-int.h>

typedef struct {
	gboolean running;
	gboolean working;

	GPid pid;
	int in, out;

	guint32 last_packet;
	gint synchro;

	GString *buf;
	wtap_dumper *pdh;
	emem_tree_t *tree;
} snort_session_t;

typedef struct {
	/* time */
	struct timeval tv;
	/* rule */
	unsigned long gen, sid, rev;

	char *msg;
	char *klass;
	int prio;
} snort_alert_t;

static int proto_snort = -1;

static int hf_snort_class = -1;
static int hf_snort_msg = -1;
static int hf_snort_rev = -1;
static int hf_snort_sid = -1;
static int hf_snort_generator = -1;
static int hf_snort_priority = -1;

static int ett_snort = -1;

static guint pref_snort_config_sync = 100;
static const char *pref_snort_config_filename = "/etc/snort/snort.conf";

static snort_session_t current_session;
static int snort_config_ok;

static void
snort_reaper(GPid pid, gint status, gpointer data)
{
	snort_session_t *session = (snort_session_t *) data;

	if (session->running && session->pid == pid) {
		g_print("snort died! %d (%d)\n", pid, status);
		session->working = session->running = FALSE;
		/* XXX, cleanup */
	} else {
		g_print("Errrrmm snort_reaper() %d != %d\n", session->pid, pid);
	}

	g_spawn_close_pid(pid);
}

static const char *
snort_parse_ts(const char *ts, struct timeval *tv)
{
	struct tm tm;
	unsigned int usec;
	
	/* timestamp */
	memset(&tm, 0, sizeof(tm));
	tm.tm_isdst = -1;
	if (sscanf(ts, "%02d/%02d/%02d-%02d:%02d:%02d.%06u ", &(tm.tm_mon), &(tm.tm_mday), &(tm.tm_year), &(tm.tm_hour), &(tm.tm_min), &(tm.tm_sec), &usec) != 7) 
		return NULL;
	tm.tm_mon -= 1; tm.tm_year += 100;

	tv->tv_sec = mktime(&tm);
	tv->tv_usec = usec;

	return strchr(ts, ' ');
}

/**
 * snort_parse_fast_line()
 *
 * Try to parse line generated from spo_alert_fast (-A console)
 */
static gboolean
snort_parse_fast_line(const char *line, snort_alert_t *al)
{
	static const char drop[] = " [Drop]";
	static const char wdrop[] = " [WDrop]";
	static const char stars[] = " [**] ";

	static const char classification[] = "[Classification: ";
	static const char priority[] = "[Priority: ";

 /* TS = %02d/%02d/%02d-%02d:%02d:%02d.%06u ", lt->tm_mon + 1, lt->tm_mday, lt->tm_year - 100, s / 3600, (s % 3600) / 60, s % 60, (u_int) tvp->tv_usec); 
  * DROPPED = " [Drop]" | " [WDrop]"
  */
	const char *tmp_msg;

	if (!(line = snort_parse_ts(line, &(al->tv))))
		return FALSE;

	/* [**] */
	if (!g_str_has_prefix(line+1, stars))
		return FALSE;
	line += sizeof(stars);

	/* [%lu:%lu:%lu] */
	if (sscanf(line, "[%lu:%lu:%lu] ", &(al->gen), &(al->sid), &(al->rev)) != 3)
		return FALSE;
	if (!(line = strchr(line, ' ')))
		return FALSE;

	tmp_msg = line+1;
	if (!(line = strstr(line, stars)))
		return FALSE;

	al->msg = se_strndup(tmp_msg, line - tmp_msg);
	line += (sizeof(stars)-1);

	/* [Classification: Attempted Administrator Privilege Gain] [Priority: 10] */

	if (g_str_has_prefix(line, classification)) {
		/* [Classification: %s] */
		char *tmp;
		line += (sizeof(classification)-1);

		if (!(tmp = strstr(line, "] [Priority: ")))
			return FALSE;

		/* assume "] [Priority: " is not inside classification text :) */
		al->klass = se_strndup(line, tmp - line);

		line = tmp+2;
	} else
		al->klass = NULL;

	/* Optimized: if al->klass we already checked this in strstr() above */
	if (al->klass || g_str_has_prefix(line, priority)) {
		/* [Priority: %d] */
		line += (sizeof(priority)-1);

		if ((sscanf(line, "%d", &(al->prio))) != 1)
			return FALSE;

		if (!(line = strstr(line, "] ")))
			return FALSE;

		line += 2;
	} else
		al->prio = -1; /* XXX */

#if 0	/* nothing interesting here */
	/* {%s} (protocol) */
	/* "%s -> %s", ip1, ip2 (OR) "%s:%d -> %s:%d */
	g_print(":: %s\n", line);
#endif

	return TRUE;
}

static gboolean
snort_fast_output(GIOChannel *source, GIOCondition condition, gpointer data)
{
	snort_session_t *session = (snort_session_t *) data;

	while (condition & G_IO_IN) {
		GIOStatus status;
		char _buf[1024];
		gsize len = 0;

		char *old_buf = NULL;
		char *buf = _buf;
		char *line;

		if ((status = g_io_channel_read_chars(source, _buf, sizeof(_buf)-1, &len, NULL)) != G_IO_STATUS_NORMAL) {
			if (status == G_IO_STATUS_AGAIN) {
				condition &= ~G_IO_IN;
				break;
			}
			g_print("snort_fast_output() status: (%d)\n", status);
			return FALSE;
		}
		buf[len] = '\0';

		if (session->buf) {
			g_string_append(session->buf, buf);

			buf = old_buf = g_string_free(session->buf, FALSE);
			session->buf = NULL;
		}

		while ((line = strchr(buf, '\n'))) {
			snort_alert_t alert;
			const char *after_ts;
			struct timeval tv;

			*line = '\0';

			if (session->synchro) {
				if (session->synchro == 2 && !strcmp(buf, "=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+"))
					session->synchro = 1;

				if (session->synchro == 1 && !buf[0])
					session->synchro = 0;

			} else if (snort_parse_fast_line(buf, &alert)) {
				g_print("%ld.%lu [%lu,%lu,%lu] %s {%s} [%d]\n", 
						alert.tv.tv_sec, alert.tv.tv_usec, 
						alert.gen, alert.sid, alert.rev, 
						alert.msg, 
						alert.klass ? alert.klass : "(null)",
						alert.prio);

				se_tree_insert32(session->tree, alert.tv.tv_usec, se_memdup(&alert, sizeof(alert)));

			} else if (pref_snort_config_sync && (after_ts = snort_parse_ts(buf, &tv))) {
				session->last_packet = MAX(session->last_packet, (guint32) tv.tv_usec);

				/* g_print("snort_fast_output() sync point: %u\n", session->last_packet); */

				session->synchro = 2;

			} else
				g_print("snort_fast_output() line: '%s'\n", buf);

			buf = line+1;
		}

		if (buf[0]) {
			/* g_print("snort_fast_output() left: '%s'\n", buf); */
			session->buf = g_string_new(buf);
		}

		g_free(old_buf);
	}
	if (condition) {
		int ishup = !!(condition & G_IO_HUP);
		int iserr = !!(condition & G_IO_ERR);

		condition &= ~(G_IO_HUP|G_IO_ERR);

		g_print("snort_fast_output() cond: (h:%d,e:%d,r:%d)\n", ishup, iserr, condition);
		return FALSE;
	}
	return TRUE;
}

static void 
snort_dissector(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree) 
{
	snort_alert_t *alert;

	/* pass it to snort */
	if (pinfo->fd->flags.visited == 0 && current_session.working) {
		int write_err;
		struct wtap_pkthdr wtp;

		if (!current_session.pdh) {
			int open_err;

			/* snort don't have format to support capture file with several encapsulations (like pcapng), 
			 * so write in pcap format and hope we have just one encap */
			current_session.pdh = wtap_dump_fdopen(current_session.in, WTAP_FILE_PCAP, pinfo->fd->lnk_t, WTAP_MAX_PACKET_SIZE, /* .compressed = */FALSE, &open_err);
			if (!current_session.pdh) {
				g_print("snort-open-err: %d\n", open_err);
				current_session.working = FALSE;
				return;
			}
		}

		wtp.ts.secs = pinfo->fd->abs_ts.secs;
		wtp.ts.nsecs = pinfo->fd->abs_ts.nsecs;

		wtp.ts.nsecs = pinfo->fd->num * 1000;	/* XXX, max 999'999 frames */

		wtp.caplen = tvb_length(tvb); 		/* pinfo->fd->cap_len */
		wtp.len = tvb_reported_length(tvb);	/* pinfo->fd->pkt_len */
		wtp.pkt_encap = pinfo->fd->lnk_t;
		if (current_session.pdh->encap != wtp.pkt_encap)
			{ /* XXX, warning! comnvert? */ } 

		if (!wtap_dump(current_session.pdh, &wtp, pinfo->pseudo_header, tvb_get_ptr(tvb, 0, tvb_length(tvb)), &write_err)) {
			g_print("snort-write-err: %d\n", write_err);
			current_session.working = FALSE;
			return;
		}
		wtap_dump_flush(current_session.pdh);

		g_main_context_iteration(NULL, FALSE);

		if (pref_snort_config_sync) {
			int tries = pref_snort_config_sync;

			while (current_session.last_packet < pinfo->fd->num && tries--) {
				g_main_context_iteration(NULL, FALSE);
				g_usleep(G_USEC_PER_SEC/1000);	/* 1ms per try */
			}
			/* g_print("got sync after: %d\n", tries); */

			if (tries <= 0) {
				/* g_print("giving up %d\n", current_session.last_packet); */
				current_session.last_packet = G_MAXUINT32;
			}
		}
	}

	if (current_session.tree && (alert = se_tree_lookup32(current_session.tree, pinfo->fd->num))) {
		proto_tree *snort_tree = NULL;

		if (tree) {
			proto_item *ti = proto_tree_add_protocol_format(tree, proto_snort, tvb, 0, -1, "Snort: (msg: \"%s\" sid: %lu)", alert->msg, alert->sid);
			snort_tree = proto_item_add_subtree(ti, ett_snort);
		}

		if (alert->klass)
			proto_tree_add_string(snort_tree, hf_snort_class, tvb, 0, 0, alert->klass);

		proto_tree_add_string(snort_tree, hf_snort_msg, tvb, 0, 0, alert->msg);

		proto_tree_add_uint(snort_tree, hf_snort_sid, tvb, 0, 0, alert->sid);
		proto_tree_add_uint(snort_tree, hf_snort_generator, tvb, 0, 0, alert->gen);
		proto_tree_add_uint(snort_tree, hf_snort_rev, tvb, 0, 0, alert->rev);
		if (alert->prio != -1)
			proto_tree_add_uint(snort_tree, hf_snort_priority, tvb, 0, 0, alert->prio);
	} else {
		/* XXX, here either this frame doesn't generate alert or we haven't received data from snort (async) 
		 *
		 *      It's problem when user want to filter tree on initial run, or is running one-pass tshark.
		 */
	}
}

static int
snort_set_line_buffered(void)
{
	/* XXX, more portable way would be to use stdbuf directly,
	 *      or unbuffer (I don't have this one, but there's man page: http://linuxcommand.org/man_pages/unbuffer1.html)
	 *
	 *      or maybe just put missing setlinebuf() in snort main()?
	 */

	/* this logic is copied from sources of coreutils/stdbuf program. [man (1) stdbuf]
	   Copyright (C) 2009-2011 Free Software Foundation, Inc.
	   Written by Pdraig Brady.  */

	static const char libstdbuf[] = "/usr/lib/coreutils/libstdbuf.so";
	char *old_libs;
	char *LD_PRELOAD;

	if (!(g_file_test(libstdbuf, G_FILE_TEST_EXISTS))) {
		errno = ENOENT;
		return -1;
	}

	if (putenv("_STDBUF_O=L"))
		return -1;

	/* XXX, do we really want to keep users LD_PRELOAD?! */
	if ((old_libs = getenv("LD_PRELOAD")))
		LD_PRELOAD = g_strdup_printf("LD_PRELOAD=%s:%s", old_libs, libstdbuf);
	else
		LD_PRELOAD = g_strdup_printf("LD_PRELOAD=%s", libstdbuf);

	if (putenv(LD_PRELOAD))
		return -1;
	return 0;
}

static void
snort_config(gpointer user_data _U_)
{
	/* try to get line-buffered (or unbuffered) output from snort */

	/* nah, it won't work */
	/* setlinebuf(stdout); */

	if (snort_set_line_buffered())
	 { }
}

/* #define ALERT_OUT "console:full"
#define ALERT_OUT "console:test" */
#define ALERT_OUT "console"

static void
snort_start(void)
{
	GIOChannel *channel;
	int snort_output_id;
	const char *argv[] = {
		"snort", "-c", pref_snort_config_filename,
		/* read from stdin */
		"-r", "-",
		/* don't log */
		"-N",
		/* output to console and silence snort */
		"-A", ALERT_OUT, "-q",
		/* normalize time */
		"-y", /* -U", */
		(pref_snort_config_sync) ? "-v" : NULL,
		NULL
	};

	if (current_session.running) {
		g_print("snort.running!\n");
		return;
	}

	current_session.running = TRUE;

	if (!snort_config_ok) {
		current_session.running = FALSE;
		g_print("!snort.config!\n");
		/* report_failure("This dissection will be performed without snort!\n"); */
		return;
	}

	if (!g_spawn_async_with_pipes(NULL, (char **) argv, NULL, G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD, snort_config, NULL, &current_session.pid, &current_session.in, &current_session.out, NULL, NULL)) {
		current_session.running = FALSE;
		g_print("fail spawn!\n");
		return;
	}
	g_print("spawn OK pid: %d!\n", current_session.pid);
	g_child_watch_add(current_session.pid, snort_reaper, &current_session);

	channel = g_io_channel_unix_new(current_session.out);
	g_io_channel_set_encoding(channel, NULL, NULL);
	g_io_channel_set_buffered(channel, FALSE);
	g_io_channel_set_flags(channel, G_IO_FLAG_NONBLOCK, NULL);
	current_session.buf = NULL;
	current_session.last_packet = 0;
	snort_output_id = g_io_add_watch_full(channel,
                                                   G_PRIORITY_HIGH,
                                                   G_IO_IN|G_IO_ERR|G_IO_HUP,
                                                   snort_fast_output,
                                                   &current_session,
                                                   NULL);
	current_session.tree = se_tree_create(EMEM_TREE_TYPE_RED_BLACK, "SNORT Tree");
	current_session.working = TRUE;
}

static void
snort_eof(void)
{
	if (current_session.pdh) {
		int write_err;

		if (!wtap_dump_close(current_session.pdh, &write_err))
			{ }
		current_session.pdh = NULL;
	}
}

extern
void
proto_reg_handoff_wireshnork(void)
{
	GError *err = NULL;
	char *snort_out;
	char *snort_err;
	int snort_exit;

	/* try to keep all args which are in snort_start(). It will detect breakage on older/newer snort which doesn't support some options */
	const char *argv[] = {
		"snort", "-c", pref_snort_config_filename,
		/* self test */
		"-T",
		/* don't log */
		"-N",
		/* output to console and silence snort */
		"-A", ALERT_OUT, "-q", /* XXX, -q quiets-out some interesting output... */
		/* normalize time */
		"-y", /* -U", */
		(pref_snort_config_sync) ? "-v" : NULL,
		NULL
	};

	g_print("proto_reg_handoff_wireshnork()\n");

	if (current_session.running) {
		g_print("snort.running!\n");
		/* XXX, configuration changed when snort is running WHAT TO DO?! */
		/* return; */ /* well we can test if config is ok even if is running. It won't just effect current running snort... */

		/* XXX, what if admin changes config when wireshark is running. Should we have inotify on pref_snort_config_filename? :> */
	}

	if (!pref_snort_config_filename[0])
		return;

	/* perform self test */
	if (!g_spawn_sync(NULL, (char **) argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, &snort_out, &snort_err, &snort_exit, &err)) {
		snort_config_ok = FALSE;
		report_failure("Snort self test failed, check if you have snort installed and it's in your $PATH\n\nError: %s\n", err->message);
		g_error_free(err);
		return;
	}

	snort_config_ok = (snort_exit == 0);

	/* g_print("snort-test: %d (%s) [%s]\n", snort_exit, snort_out, snort_err); */
	if (!snort_config_ok) {
		static const char fatal_error[] = "Fatal Error, Quitting..\n";

		/* Truncate message */
		if (g_str_has_suffix(snort_err, fatal_error))
			snort_err[strlen(snort_err)-strlen(fatal_error)] = '\0';

		report_failure("Snort self test failed (exit code: %d)\n\nstdout: %s\n\nstderr: %s\n", snort_exit, snort_out, snort_err);
	}

	g_free(snort_out);
	g_free(snort_err);
}

extern
void
proto_register_wireshnork(void)
{
	static hf_register_info hf[] = {
		{ &hf_snort_sid, 
			{ "Rule SID", "snort.sid", FT_UINT32, BASE_DEC, NULL, 0x00,
			NULL, HFILL }},

		{ &hf_snort_msg,
			{ "Alert Message", "snort.msg", FT_STRINGZ, BASE_NONE, NULL, 0x00,
			NULL, HFILL }},
		{ &hf_snort_class,
			{ "Alert Classification", "snort.class", FT_STRINGZ, BASE_NONE, NULL, 0x00,
			NULL, HFILL }},
		{ &hf_snort_priority,
			{ "Alert Priority", "snort.priority", FT_UINT32, BASE_DEC, NULL, 0x00,
			NULL, HFILL }},
		{ &hf_snort_generator,
			{ "Rule Generator", "snort.generator", FT_UINT32, BASE_DEC, NULL, 0x00,
			NULL, HFILL }},
		{ &hf_snort_rev,
			{ "Rule Revision", "snort.rev", FT_UINT32, BASE_DEC, NULL, 0x00,
			NULL, HFILL }},
	};
	static gint *ett[] = {
		&ett_snort
	};

	/* static enum_val_t snort_output_vals[] = {
		{ "output-test", "Output rule (NOT IMPLEMENTED)", SNORT_OUTPUT_TEST },
		{ "output-fast", "Output rule, msg, class, priority (Limited to 1'000'000 packets)", SNORT_OUTPUT_FAST },
		{ "output-full", "Output rule, msg, class, priority, references (NOT IMPLEMENTED)", SNORT_OUTPUT_FULL }, 
		{ NULL, NULL, -1}
	}; */

	dissector_handle_t snort_handle;
	module_t *snort_module;

	proto_snort = proto_register_protocol("Wireshark Snort Plugin", "WireShnork", "snort");

	g_print("proto_register_wireshnork()\n");
	proto_register_field_array(proto_snort, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	snort_module = prefs_register_protocol(proto_snort, proto_reg_handoff_wireshnork);
	prefs_register_string_preference(snort_module, "config",
					 "Configuration Filename",
					 "The name of the file containing the snort IDS configuration",
					 &pref_snort_config_filename);
	prefs_register_uint_preference(snort_module, "sync",
					"Try to synchronize communication with snort [ms]" ,
					"This option pass extra -v (verbose) option to snort. This option makes snort and WireShnort plugin a little slower. Enabling this option is *REQUIRED* for one pass tshark",
					10, 
					&pref_snort_config_sync);

	snort_handle = create_dissector_handle(snort_dissector, proto_snort);

	register_init_routine(snort_start);
	register_postdissector(snort_handle);
	register_postseq_cleanup_routine(snort_eof);
}

