/* packet-wiresocks.c
 *
 * Copyright 2011, Jakub Zawadzki <darkjames-ws@darkjames.pl>
 *
 * Google Summer of Code 2011 for The Honeynet Project
 * Mentors:
 *    Guillaume Arcas <guillaume.arcas (at) retiaire.org>
 *    Jeff Nathan <jeffnathan (at) gmail.com>
 *
 * $Id$
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <errno.h>
#include <glib.h>

#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <epan/epan_dissect.h>
#include <epan/epan.h>
#include <epan/packet.h>
#include <epan/pint.h>
#include <epan/prefs.h>
#include <epan/reassemble.h>
#include <epan/etypes.h>
#include <epan/expert.h>
#include <epan/report_err.h>
#include <epan/tap.h>
#include <epan/tfs.h>

#include <wiretap/wtap.h>

typedef struct {
	GString *input;
	GString *output;

} proxy_client_t;

typedef struct {
	frame_data *fdata;

	char *url;
	char *type;
} http_file_t;

static struct wtap *current_wth;

static emem_tree_t empty_tree;
static emem_tree_t *http_files_tree = &empty_tree;

#include <epan/dissectors/packet-http.h>
static int
tap_http_packet(void *tapdata _U_, packet_info *pinfo, epan_dissect_t *edt _U_, const void *data)
{
	const http_eo_t *eo_info = (const http_eo_t *) data;

	if (eo_info) {
		if (!eo_info->hostname || !eo_info->filename) {
			g_print("XXX1: %s %s\n", eo_info->hostname, eo_info->filename);
			return 0;
		}

		if (!se_tree_lookup32(http_files_tree, pinfo->fd->num)) {
			http_file_t *file = se_alloc0(sizeof(http_file_t));

			if (pinfo->srcport != 80)
				file->url = se_strdup_printf("http://%s:%u%s", eo_info->hostname, pinfo->srcport, eo_info->filename);
			else
				file->url = se_strdup_printf("http://%s%s", eo_info->hostname, eo_info->filename);
			file->type = se_strdup(eo_info->content_type);
			file->fdata = pinfo->fd;

			se_tree_insert32(http_files_tree, pinfo->fd->num, file);
		}
#if 0
		guint32 addr;

		if (pinfo->srcport != 80 && pinfo->destport == 80) {
			g_print("XXX2: %u %u\n", pinfo->srcport, pinfo->destport);
			return 0;
		}

		if (pinfo->dst.type != AT_IPv4 || pinfo->dst.len != 4) {
			g_print("XXX3: %u %d\n", pinfo->dst.type, pinfo->dst.len);
			return 0;
		}
		memcpy(&addr, pinfo->dst.data, 4);

		if (!se_tree_lookup32(http_hosts_tree, addr)) { /* ... */ }
#endif
	}
	return 0;
}

typedef struct {
	const char *url;

	frame_data *result;
} find_frame_t;

static gboolean
http_find_frame_cb(void *value, void *userdata)
{
	const http_file_t *file = (const http_file_t *) value;
	find_frame_t *data = (find_frame_t *) userdata;

	if (!strcmp(data->url, file->url)) {
		data->result = file->fdata;
		return TRUE;
	}

	return FALSE;
}

static frame_data *
http_find_frame(const char *url)
{
	find_frame_t data;

	data.url = url;
	data.result = NULL;

	emem_tree_foreach(http_files_tree, http_find_frame_cb, &data);
	return data.result;
}

static void
socks_start(void)
{
	http_files_tree = se_tree_create(EMEM_TREE_TYPE_RED_BLACK, "HTTP Files Tree");
}

static int
server_up(const char *ip, int port)
{
	struct sockaddr_in sin;
	struct sockaddr_in6 sin6;

	struct sockaddr *sa;
	socklen_t salen;

	int fd;
	int err;
	int opt = 1;

	if (strchr(ip, ':')) {
		sin6.sin6_family = AF_INET6;
		sin6.sin6_port = htons(port);
		if (inet_pton(AF_INET6, ip, &sin6.sin6_addr) <= 0) {
			errno = EINVAL;
			return -1;
		}

		sa = (struct sockaddr *) &sin6;
		salen = sizeof(sin);
	} else {
		sin.sin_family = AF_INET;
		sin.sin_port = htons(port);
		if (!inet_aton(ip, &sin.sin_addr))
			return -1;

		sa = (struct sockaddr *) &sin;
		salen = sizeof(sin);
	}

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		return -1;

	/* not fatal */
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));

	if (bind(fd, sa, salen))
		goto err;

	if (listen(fd, SOMAXCONN))
		goto err;
	
	return fd;
err:
	err = errno;
	close(fd);
	errno = err;
	return -1;
}

#define writes(fd, str) write(fd, str, sizeof(str)-1)

static int
http_write_file_cb(void *tapdata _U_, packet_info *pinfo _U_, epan_dissect_t *edt _U_, const void *data)
{
	const http_eo_t *eo_info = (const http_eo_t *) data;
	int fd = GPOINTER_TO_INT(tapdata);

	if (eo_info) {
		if (eo_info->content_type) {
			writes(fd, "Content-type: "); write(fd, eo_info->content_type, strlen(eo_info->content_type)); writes(fd, "\r\n");
		}
		/* write(fd, "Content-length: %lu\r\n", eo_info->payload_len); */

		writes(fd, "\r\n");
		if (eo_info->payload_data)
			write(fd, eo_info->payload_data, eo_info->payload_len);
	}
	return 0;
}

static gboolean
http_list_socket_cb(void *value, void *userdata)
{
	const http_file_t *file = value;
	int fd = GPOINTER_TO_INT(userdata);
	char buf[16];

	g_snprintf(buf, sizeof(buf), "'/%u'>", file->fdata->num);

	writes(fd, "<li><a href=");
		write(fd, buf, strlen(buf));
		write(fd, file->url, strlen(file->url));
	writes(fd, "</a></li>\n");

	return FALSE;
}

static void
http_handle_request(int fd, const char *url)
{
	frame_data *fdata = NULL;

#define STATUS_200_OK        "HTTP/1.0 200 OK\r\n"
#define STATUS_301_MOVED     "HTTP/1.0 301 Moved\r\n"
#define STATUS_404_NOT_FOUND "HTTP/1.0 404 Not Found\r\n"

	if (url[0] == '/') {
		unsigned int no = -1;

		if (url[1] == '\0') {
			writes(fd, STATUS_200_OK);
			writes(fd, "\r\n");

			writes(fd, "<html><body><ul>");
			emem_tree_foreach(http_files_tree, http_list_socket_cb, GINT_TO_POINTER(fd));
			writes(fd, "</ul></body>");
			return;

		} else if (sscanf(url + 1, "%u", &no) == 1) {
			const http_file_t *file = se_tree_lookup32(http_files_tree, no);

			if (file) {
				fdata = file->fdata;


			}
		}

	} else
		fdata = http_find_frame(url);

	if (fdata) {
		union wtap_pseudo_header pseudo_header;
		guint8 pd[WTAP_MAX_PACKET_SIZE+1];
		int err;
		gchar *err_info;

		if (wtap_seek_read(current_wth, fdata->file_off, &pseudo_header, pd, fdata->cap_len, &err, &err_info)) {
			epan_dissect_t edt;

			writes(fd, STATUS_200_OK);

			register_tap_listener("http_eo", GINT_TO_POINTER(fd), NULL, TL_IS_DISSECTOR_HELPER, NULL, http_write_file_cb, NULL);

			epan_dissect_init(&edt, FALSE, FALSE);
			tap_queue_init(&edt);
			epan_dissect_run(&edt, &pseudo_header, pd, fdata, NULL);
			tap_push_tapped_queue(&edt);
			epan_dissect_cleanup(&edt);

			remove_tap_listener(GINT_TO_POINTER(fd));

		} else
			g_free(err_info);

	} else {
		writes(fd, STATUS_404_NOT_FOUND);
		writes(fd, "\r\n");
		writes(fd, STATUS_404_NOT_FOUND);
	}
}

static gboolean
http_read_socket(GIOChannel *io, GIOCondition cond, void *data)
{
	proxy_client_t *cli = (proxy_client_t *) data;
	int fd = g_io_channel_unix_get_fd(io);

	char _buf[1024+1];
	char *buf;
	int ret;

	if (cond & G_IO_IN) {
		ret = read(fd, _buf, sizeof(_buf)-1);
		if (ret <= 0)
			return FALSE;

		if (cli->input->len) {
			g_string_append_len(cli->input, _buf, ret);
			buf = cli->input->str;
		} else {
			buf = _buf;
			_buf[ret] = '\0';
		}

	} else
		buf = cli->input->str;

	g_print("http: %s\n", buf);
	if (g_str_has_prefix(buf, "GET ")) {
		char *url = buf + 4;
		char *tmp;

		if ((tmp = strchr(url, ' ')))
			*tmp = '\0';

		http_handle_request(fd, url);
		return FALSE;
	}
	return TRUE;
}

static gboolean
socks_http_read_socket(GIOChannel *io, GIOCondition cond, void *data)
{
	proxy_client_t *cli = (proxy_client_t *) data;
	int fd = g_io_channel_unix_get_fd(io);

	char _buf[1024+1];
	char *buf;
	int ret;

	if (cond & G_IO_IN) {
		ret = read(fd, _buf, sizeof(_buf)-1);
		if (ret <= 0)
			return FALSE;

		if (cli->input->len) {
			g_string_append_len(cli->input, _buf, ret);
			buf = cli->input->str;
		} else {
			buf = _buf;
			_buf[ret] = '\0';
		}

	} else
		buf = cli->input->str;

	g_print("socks: %s\n", buf);
	if ((strstr(buf, "\r\n\r\n"))) {
		if (g_str_has_prefix(buf, "GET ")) {
			char *file = buf + 4;
			char *host = NULL;
			char *tmp;

			if ((tmp = strchr(file, ' '))) {
				*tmp = '\0';
				buf = tmp+1;
			}

			if ((tmp = strstr(buf, "\r\nHost: "))) {
				host = tmp + 8;
				if ((tmp = strstr(host, "\r\n")))
					*tmp = '\0';
			}

			if (host && file) {
				char *url = g_strdup_printf("http://%s%s", host, file);
				http_handle_request(fd, url);
				g_free(url);
			}
		}
		return FALSE;
	}
	return TRUE;
}

static gboolean
socks_read_socket(GIOChannel *io, GIOCondition cond, void *data)
{
	static const char http[] = { 'G', 'E', 'T', ' ' };

	proxy_client_t *cli = (proxy_client_t *) data;
	int fd = g_io_channel_unix_get_fd(io);

	guint8 buf[1024];
	int i, ret;

	ret = read(fd, buf, sizeof(buf));
	if (ret <= 0)
		return FALSE;

	if (ret >= sizeof(http) && !memcmp(buf, http, sizeof(http))) {
		/* pass to HTTP */
		g_string_append_len(cli->input, buf, ret);
		if (http_read_socket(io, 0, data))
			g_io_add_watch(io, G_IO_IN, http_read_socket, data);
		return FALSE;
	}

	if (buf[0] == 4) { /* # field 1: SOCKS version number, 1 byte, must be 0x04 for this version */
		guint8 cmd_code = buf[1];		/* field 2: command code, 1 byte:  */
#if 0
		guint16 port = pntohs(&buf[2]); /* field 3: network byte order port number, 2 bytes */
		guint32 addr = pntohl(&buf[4]); /* field 4: network byte order IP address, 4 bytes */

		/* Socks4A */
		if ((addr & 0xffffff00) == 0 && (addr & 0x000000ff) != 0) {

		}
#endif

		/* 0x01 = establish a TCP/IP stream connection */
		if (cmd_code == 0x01 /* && port == 80 */) {
			static const char socks4_ok[] = { 0x00, 0x5a, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD  };

			write(fd, socks4_ok, sizeof(socks4_ok));

			g_io_add_watch(io, G_IO_IN, socks_http_read_socket, data);
			return FALSE;

		} else {
			static const char socks4_rej[] = { 0x00, 0x5b, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD  };

			write(fd, socks4_rej, sizeof(socks4_rej));
			return FALSE;
		}

	} else if (buf[0] == 5) {	/* field 1: SOCKS version number (must be 0x05 for this version */

	}

	printf("Unhandled: ");
	for (i = 0; i < ret; i++)
		printf("%.2x ", buf[i]);
	printf("\n");

	return FALSE;
}

static gboolean
socks_accept_socket(GIOChannel *io, GIOCondition cond, void *data _U_)
{
	int sfd = g_io_channel_unix_get_fd(io);

	struct sockaddr_in sin;
	socklen_t sinlen;
	int fd;
	
	if (cond & G_IO_IN) {
		GIOChannel *io;

		/* XXX, IPv6 */
		fd = accept(sfd, (struct sockaddr *) &sin, &sinlen);

		g_print("socks_accept_socket(): %d\n", fd);

		if (fd != -1) {
			proxy_client_t *client = g_malloc(sizeof(proxy_client_t));

			client->input = g_string_new(NULL);
			client->output = g_string_new(NULL);

			io = g_io_channel_unix_new(fd);
			g_io_channel_set_flags(io, G_IO_FLAG_NONBLOCK, NULL);
			g_io_channel_set_close_on_unref(io, TRUE);
			/* for now let's assume it's socks connection */
			g_io_add_watch(io, G_IO_IN, socks_read_socket, client);

			g_io_channel_unref(io);
		}
	}
	return TRUE;
}

extern
void
proto_reg_handoff_wiresocks(void)
{
	GIOChannel *io;
	int fd;
	
	fd = server_up("127.0.0.1", 3128);

	if (fd == -1) {
		report_failure("proto_reg_handoff_wiresocks() server up fail: %s\n", g_strerror(errno));
		return;
	}

	io = g_io_channel_unix_new(fd);
	g_io_channel_set_close_on_unref(io, TRUE);
	g_io_add_watch(io, G_IO_IN, socks_accept_socket, NULL);
	g_io_channel_unref(io);

	register_tap_listener("http_eo", NULL, NULL, TL_IS_DISSECTOR_HELPER, NULL, tap_http_packet, NULL);
}

extern
void
proto_register_wiresocks(void)
{
	g_print("proto_register_wiresocks()\n");
	register_init_routine(socks_start);
}


/* Hack,
 *    we're registering wtap open routine to get wtap handle ;-)
 */
static gint
wiresocks_wtap_open(struct wtap *wtap, int *err _U_, char **err_info _U_)
{
	g_print("wiresocks_wtap_open(): %p\n", wtap);
	current_wth = wtap;
	return 0;
}

extern
void 
wtap_register_wiresocks_helper(void) {
	wtap_register_open_routine(wiresocks_wtap_open, TRUE);
}

