/* conv.c
 *
 * Copyright 2011, Jakub Zawadzki <darkjames-ws@darkjames.pl>
 *
 * Google Summer of Code 2011 for The Honeynet Project
 * Mentors:
 *    Guillaume Arcas <guillaume.arcas (at) retiaire.org>
 *    Jeff Nathan <jeffnathan (at) gmail.com>
 *
 * $Id$
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>

#include <globals.h>

#include <epan/packet.h>
#include <epan/tap.h>

#include "conv.h"

typedef struct {
	GHashTable *table;

	void (*func)(conv_data_t *, epan_dissect_t *, gpointer);
	void *data;
} conv_cb_t;

static void
conv_common_cb(conv_cb_t *conv, const packet_info *pinfo, epan_dissect_t *edt, const address *src, const address *dst, guint32 src_port, guint32 dst_port, int port_type_val)
{
	conv_data_t *conversation;
	conv_key_t existing_key;
	int cmp_ret;

	cmp_ret = CMP_ADDRESS(src, dst);
	if (cmp_ret == 0)
		cmp_ret = (src_port > dst_port) ? 1 : (src_port < dst_port) ? -1 : 0;

	if (cmp_ret > 0) {
		const address *tmp_a = src;
		guint32 tmp_p = src_port;

		src = dst;
		src_port = dst_port;
		dst = tmp_a;
		dst_port = tmp_p;
    }

	existing_key.src_address = *src;
	existing_key.dst_address = *dst;
	existing_key.src_port = src_port;
	existing_key.dst_port = dst_port;

	/* try to find it among the existing known conversations */
	if (!(conversation = g_hash_table_lookup(conv->table, &existing_key))) {
		/* if we still dont know what conversation this is it has to be a new one
		   and we have to allocate it and append it to the end of the list */
		conversation = g_malloc(sizeof(conv_data_t));

		COPY_ADDRESS(&conversation->key.src_address, src);
		COPY_ADDRESS(&conversation->key.dst_address, dst);
		conversation->key.src_port = src_port;
		conversation->key.dst_port = dst_port;

		conversation->port_type = port_type_val;
		conversation->rx_frames = 0;
		conversation->tx_frames = 0;
		conversation->rx_bytes = 0;
		conversation->tx_bytes = 0;

		conversation->start_time = pinfo->fd->rel_ts;
		conversation->stop_time = pinfo->fd->rel_ts;

		g_hash_table_insert(conv->table, &(conversation->key), conversation);
	}

	/* update the conversation struct */
	if (cmp_ret > 0) {
		conversation->rx_frames += 1;
		conversation->rx_bytes += pinfo->fd->pkt_len;
	} else {
		conversation->tx_frames += 1;
		conversation->tx_bytes += pinfo->fd->pkt_len;
	}

	if (nstime_cmp(&pinfo->fd->rel_ts, &conversation->stop_time) > 0)
		conversation->stop_time = pinfo->fd->rel_ts;
	else if (nstime_cmp(&pinfo->fd->rel_ts, &conversation->start_time) < 0)
		conversation->start_time = pinfo->fd->rel_ts;

	if (conv->func)
		conv->func(conversation, edt, conv->data);
}

#include <epan/dissectors/packet-eth.h>
static int
conv_eth_cb(void *tapdata, packet_info *pinfo, epan_dissect_t *edt, const void *data)
{
	const eth_hdr *ehdr = (const eth_hdr *) data;
	conv_cb_t *conv = (conv_cb_t *) tapdata;

	conv_common_cb(conv, pinfo, edt, &ehdr->src, &ehdr->dst, 0, 0, PT_NONE);
	return 0;
}

#include <epan/dissectors/packet-ip.h>
static int
conv_ip_cb(void *tapdata, packet_info *pinfo, epan_dissect_t *edt, const void *data)
{
	const ws_ip *iph = (const ws_ip *) data;
	conv_cb_t *conv = (conv_cb_t *) tapdata;

	conv_common_cb(conv, pinfo, edt, &iph->ip_src, &iph->ip_dst, 0, 0, PT_NONE);
	return 0;
}

#include <epan/dissectors/packet-ipv6.h>
static int
conv_ip6_cb(void *tapdata, packet_info *pinfo, epan_dissect_t *edt, const void *data)
{
	const struct ip6_hdr *ip6h = (const struct ip6_hdr *) data;
	conv_cb_t *conv = (conv_cb_t *) tapdata;
	address src;
	address dst;

	/* Addresses aren't implemented as 'address' type in struct ip6_hdr */
	src.type = dst.type = AT_IPv6;
	src.len  = dst.len = sizeof(struct e_in6_addr);
	src.data = &ip6h->ip6_src;
	dst.data = &ip6h->ip6_dst;

	conv_common_cb(conv, pinfo, edt, &src, &dst, 0, 0, PT_NONE);
	return 0;
}

#include <epan/dissectors/packet-tcp.h>
static int
conv_tcp_cb(void *tapdata, packet_info *pinfo, epan_dissect_t *edt, const void *data)
{
	const struct tcpheader *tcphdr = (const struct tcpheader *) data;
	conv_cb_t *conv = (conv_cb_t *) tapdata;

	conv_common_cb(conv, pinfo, edt, &tcphdr->ip_src, &tcphdr->ip_dst, tcphdr->th_sport, tcphdr->th_dport, PT_TCP);
	return 0;
}

#include <epan/dissectors/packet-udp.h>
static int
conv_udp_cb(void *tapdata, packet_info *pinfo, epan_dissect_t *edt, const void *data)
{
	const e_udphdr *udphdr = (const e_udphdr *) data;
	conv_cb_t *conv = (conv_cb_t *) tapdata;

	conv_common_cb(conv, pinfo, edt, &udphdr->ip_src, &udphdr->ip_dst, udphdr->uh_sport, udphdr->uh_dport, PT_UDP);
	return 0;
}

static guint
conversation_hash(gconstpointer v)
{
	const conv_key_t *key = (const conv_key_t *)v;
	guint hash_val = 0;

	ADD_ADDRESS_TO_HASH(hash_val, &key->src_address);
	hash_val += key->src_port;
	ADD_ADDRESS_TO_HASH(hash_val, &key->dst_address);
	hash_val += key->dst_port;

	return hash_val;
}

/*
 * Compare two conversation keys for an exact match.
 */
static gboolean
conversation_match(gconstpointer v, gconstpointer w)
{
	const conv_key_t *v1 = (const conv_key_t *)v;
	const conv_key_t *v2 = (const conv_key_t *)w;

	return (v1->src_port == v2->src_port &&
			v1->dst_port == v2->dst_port &&
			ADDRESSES_EQUAL(&v1->src_address, &v2->src_address) &&
			ADDRESSES_EQUAL(&v1->dst_address, &v2->dst_address));
}

static void
conversation_free(gpointer data)
{
	conv_data_t *conv = data;

	g_free((gpointer)conv->key.src_address.data);
	g_free((gpointer)conv->key.dst_address.data);
	g_free(conv);
}

void
conv_list(const char *tap, const char *filter, GHFunc func, void (*user_func)(conv_data_t *, epan_dissect_t *, gpointer), void *user_data)
{
	tap_packet_cb packet_func;
	gboolean show_ports = FALSE;
	conv_cb_t data;

	GHashTable *hashtable;
	struct timeval tv1, tv2;

	if (!strcmp(tap, "tcp")) {
		packet_func = conv_tcp_cb;
		show_ports = TRUE;
	} else if (!strcmp(tap, "udp")) {
		packet_func = conv_udp_cb;
		show_ports = TRUE;
	} else if (!strcmp(tap, "eth")) {
		packet_func = conv_eth_cb;
	} else if (!strcmp(tap, "ip")) {
		packet_func = conv_ip_cb;
	} else if (!strcmp(tap, "ipv6")) {
		packet_func = conv_ip6_cb;
	} else
		return;
		
	data.table = hashtable = g_hash_table_new_full(conversation_hash, conversation_match, NULL, conversation_free);
	data.func = user_func;
	data.data = user_data;

	register_tap_listener(tap, &data, filter, TL_REQUIRES_PROTO_TREE/* TL_IS_DISSECTOR_HELPER */, NULL, packet_func, NULL);
gettimeofday(&tv1, NULL);
	cf_retap_packets(&cfile);
gettimeofday(&tv2, NULL);
g_print("Retap took: %lu ms\n", (tv2.tv_sec - tv1.tv_sec) * 1000 + (tv2.tv_usec - tv1.tv_usec)/1000);
	remove_tap_listener(&data);

	g_hash_table_foreach(hashtable, func, user_data);

	g_hash_table_destroy(hashtable);
}

