/**********************************************************
*      This software is part of the graphviz package      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2004 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
**********************************************************/

/* It's based on sources of graphviz gtk plugin sources
 * Cleanup by Jakub Zawadzki 2010,2011 for easier embedded into gtk applications */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <graphviz/gvc.h>
#include <graphviz/gvplugin_device.h>

#include <gtk/keys.h>
#include <gtk/gtkglobals.h>
#include <globals.h>

static void 
gtk_initialize(GVJ_t *firstjob)
{
	Display *dpy;
	int scr;

	dpy = XOpenDisplay(NULL);
	if (dpy == NULL) {
		fprintf(stderr, "Failed to open XLIB display: %s\n",
				XDisplayName(NULL));
		return;
	}
	/* dpy = gdk_x11_get_default_xdisplay(); */
	scr = DefaultScreen(dpy);
	firstjob->device_dpi.x = DisplayWidth(dpy, scr) * 25.4 / DisplayWidthMM(dpy, scr);
	firstjob->device_dpi.y = DisplayHeight(dpy, scr) * 25.4 / DisplayHeightMM(dpy, scr);
	firstjob->device_sets_dpi = TRUE;
}

static gboolean
drawingarea_scroll_event_cb(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
	const GdkEventScroll *ev = (const GdkEventScroll *) event;
	GVJ_t *job = (GVJ_t *) user_data;

	pointf pointer;
	pointer.x = ev->x;
	pointer.y = ev->y;

	switch (ev->direction) {
		case GDK_SCROLL_UP:
			(job->callbacks->button_press)(job, 4, pointer);
			break;
		case GDK_SCROLL_DOWN:
			(job->callbacks->button_press)(job, 5, pointer);
			break;
	}
	gtk_widget_queue_draw(widget);

	return FALSE;
}

static gboolean
drawingarea_configure_event_cb(GtkWidget *widget _U_, GdkEventConfigure *event, gpointer user_data)
{
	GVJ_t *job = (GVJ_t *) user_data;

	double zoom_to_fit = MIN((double) event->width / (double) job->width,
			(double) event->height / (double) job->height);

	if (!job->has_been_rendered) {
		if (zoom_to_fit < 1.0) /* don't make bigger */
			job->zoom *= zoom_to_fit;
	} else if (job->fit_mode)
		job->zoom *= zoom_to_fit;
	if (event->width > job->width || event->height > job->height)
		job->has_grown = TRUE;
	job->width = event->width;
	job->height = event->height;
	job->needs_refresh = TRUE;

	return FALSE;
}

static gboolean
drawingarea_expose_event_cb(GtkWidget *widget, GdkEventExpose *event _U_, gpointer user_data)
{
	GVJ_t *job = (GVJ_t *) user_data;
	cairo_t *cr;

	cr = gdk_cairo_create(widget->window);

	(job->callbacks->motion)(job, job->pointer);

	job->context = (void *)cr;
	job->external_context = TRUE;
	job->width = widget->allocation.width;
	job->height = widget->allocation.height;

	(job->callbacks->refresh)(job);

	cairo_destroy(cr);
	return FALSE;
}

static gboolean
drawingarea_motion_notify_event_cb(GtkWidget *widget, GdkEventMotion *event, gpointer user_data)
{
	GVJ_t *job = (GVJ_t *) user_data;
	GString *str = NULL;

	job->pointer.x = event->x;
	job->pointer.y = event->y;

	gtk_widget_queue_draw(widget);

	if (job->current_obj) {
		str = g_string_new(agget(job->current_obj, "tooltip"));
#if 0
		switch (agobjkind(job->current_obj)) {
		case AGNODE: {
			Agnode_t *n = job->current_obj;
			Agraph_t *g = n->graph;
			Agnode_t *next;
			Agedge_t *ep;

			g_string_append(str, "\nConnection list:");
			for (ep = agfstedge(g, n); ep; ep = agnxtedge(g, ep, n)) {
				if ((next = agtail(ep)) == n)
					next = aghead(ep);

				/* tx/rx bytes? time? other statistics? */
				g_string_append_printf(str, "\n- %s", agget(next, "label"));
			}
			break;
		}

		case AGEDGE:
		case AGGRAPH:
			break;
		}
#endif
	}

	gtk_widget_set_tooltip_text(widget, str ? str->str : NULL);
	gtk_widget_trigger_tooltip_query(widget);

	if (str)
		g_string_free(str, TRUE);
	return FALSE;
}

static void
prepare_filter_cb(GtkWidget *w _U_, gpointer user_data)
{
	void *obj = user_data;

	if (obj) {
		const char *filter = agget(obj, "filter");

		if (filter) {
			GtkWidget *filter_cm = g_object_get_data(G_OBJECT(top_level), E_DFILTER_CM_KEY);
			GtkWidget *filter_te = gtk_bin_get_child(GTK_BIN(filter_cm));

			gtk_entry_set_text(GTK_ENTRY(filter_te), filter);
		}
		g_print("prepare_filter_cb: %s\n", filter);
	}
}

static void
apply_filter_cb(GtkWidget *w _U_, gpointer user_data)
{
	void *obj = user_data;

	if (obj) {
		const char *filter = agget(obj, "filter");

		if (filter) {
			GtkWidget *filter_cm = g_object_get_data(G_OBJECT(top_level), E_DFILTER_CM_KEY);
			GtkWidget *filter_te = gtk_bin_get_child(GTK_BIN(filter_cm));

			gtk_entry_set_text(GTK_ENTRY(filter_te), filter);
			main_filter_packets(&cfile, filter, TRUE);
		}
		g_print("apply_filter_cb: %s\n", filter);
	}
}

static gboolean
drawingarea_button_press_event_cb(GtkWidget *widget _U_, GdkEventButton *event, gpointer user_data)
{
	GVJ_t *job = (GVJ_t *) user_data;
	pointf pointer;

	pointer.x = event->x;
	pointer.y = event->y;
	(job->callbacks->button_press)(job, event->button, pointer);

	if (event->button == 1)
		apply_filter_cb(widget, job->current_obj);

	if (event->button == 3) {
		if (job->current_obj) {
			const char *filter = agget(job->current_obj, "filter");
			GtkWidget *menu, *submenu, *item;

			if (filter)
				menu = gtk_menu_new();

			if (filter) {
				item = gtk_menu_item_new_with_label(filter);
				gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
				gtk_widget_show(item);

				submenu = gtk_menu_new();
				gtk_menu_item_set_submenu(item, submenu);
				gtk_widget_show(submenu);

				item = gtk_menu_item_new_with_label("Prepare filter");
				g_signal_connect(item, "activate", G_CALLBACK(prepare_filter_cb), job->current_obj);
				gtk_menu_shell_append(GTK_MENU_SHELL(submenu), item);
				gtk_widget_show(item);

				item = gtk_menu_item_new_with_label("Apply filter");
				g_signal_connect(item, "activate", G_CALLBACK(apply_filter_cb), job->current_obj);
				gtk_menu_shell_append(GTK_MENU_SHELL(submenu), item);
				gtk_widget_show(item);
			}

			/* XXX, statistics like RX/TX bytes */

			switch (agobjkind(job->current_obj)) {
			case AGNODE: {
				Agnode_t *n = job->current_obj;
				Agraph_t *g = n->graph;
				Agnode_t *next;
				Agedge_t *ep;

				item = gtk_menu_item_new_with_label("Connection list");
				gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
				gtk_widget_show(item);

				submenu = gtk_menu_new();
				gtk_menu_item_set_submenu(item, submenu);
				gtk_widget_show(submenu);
				for (ep = agfstedge(g, n); ep; ep = agnxtedge(g, ep, n)) {
					if ((next = agtail(ep)) == n)
						next = aghead(ep);

					/* tx/rx bytes? time? other statistics? */
					item = gtk_menu_item_new_with_label(agget(next, "label"));
					/* g_signal_connect(item, "activate", G_CALLBACK(apply_filter_cb), job); */ // XXX, apply & prepare? A<==>B
					gtk_menu_shell_append(GTK_MENU_SHELL(submenu), item);
					gtk_widget_show(item);
				}
				break;
			}

			case AGEDGE:
			case AGGRAPH:
				break;
			}


			if (filter)
				gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, event->button, event->time);
		}
	}
	return FALSE;
}

static gboolean
drawingarea_button_release_event_cb(GtkWidget *widget _U_, GdkEventButton *event, gpointer user_data)
{
	GVJ_t *job = (GVJ_t *) user_data;
	pointf pointer;

	pointer.x = event->x;
	pointer.y = event->y;
	(job->callbacks->button_release)(job, event->button, pointer);

	return FALSE;
}

static void
gtk_finalize(GVJ_t *firstjob)
{
	GtkWidget *drawing_area = (GtkWidget *) firstjob->output_file;

	gulong expose_handler = g_signal_connect(drawing_area, "expose_event", G_CALLBACK(drawingarea_expose_event_cb), firstjob);
	gulong scroll_handler = g_signal_connect(drawing_area, "scroll_event", G_CALLBACK(drawingarea_scroll_event_cb), firstjob);
	gulong notify_handler = g_signal_connect(drawing_area, "motion_notify_event", G_CALLBACK(drawingarea_motion_notify_event_cb), firstjob);
	gulong buttonp_handler = g_signal_connect(drawing_area, "button_press_event", G_CALLBACK(drawingarea_button_press_event_cb), firstjob);
	gulong buttonr_handler = g_signal_connect(drawing_area, "button_release_event", G_CALLBACK(drawingarea_button_release_event_cb), firstjob);
	gulong configure_handler = g_signal_connect(drawing_area, "configure_event", G_CALLBACK(drawingarea_configure_event_cb), firstjob);

	/* loop */
	gtk_widget_queue_draw(drawing_area);
	{
		GMainLoop *loop = g_main_loop_new(NULL, FALSE);

		g_object_set_data(G_OBJECT(drawing_area), "loop", loop);

		GDK_THREADS_LEAVE();  
		g_main_loop_run(loop);
		GDK_THREADS_ENTER();  

		g_object_set_data(G_OBJECT(drawing_area), "loop", NULL);
		g_main_loop_unref(loop);
	}

	g_signal_handler_disconnect(drawing_area, expose_handler);
	g_signal_handler_disconnect(drawing_area, scroll_handler);
	g_signal_handler_disconnect(drawing_area, notify_handler);
	g_signal_handler_disconnect(drawing_area, buttonp_handler);
	g_signal_handler_disconnect(drawing_area, buttonr_handler);
	g_signal_handler_disconnect(drawing_area, configure_handler);
}

static gvdevice_features_t device_features_gtk = {
	GVDEVICE_DOES_TRUECOLOR | GVDEVICE_EVENTS,		/* flags */
	{ 0., 0.},										/* default margin - points */
	{ 0., 0.},										/* default page width, height - points */
	{ 96., 96.},									/* dpi */
};

static gvdevice_engine_t device_engine_gtk = {
	gtk_initialize,
	NULL,			/* gtk_format */
	gtk_finalize,
};

static gvplugin_installed_t gvdevice_types_gtk[] = {
	{ 0, "gviz:cairo", 0, &device_engine_gtk, &device_features_gtk },
	{ 0, NULL, 0, NULL, NULL}
};

static gvplugin_api_t apis[] = {
	{ API_device, gvdevice_types_gtk },
	{ (api_t) 0, 0},
};

gvplugin_library_t gv_wireviz_gtk = { "gviz", apis };

