/* wireviz.c
 *
 * Copyright 2011, Jakub Zawadzki <darkjames-ws@darkjames.pl>
 *
 * Google Summer of Code 2011 for The Honeynet Project
 * Mentors:
 *    Guillaume Arcas <guillaume.arcas (at) retiaire.org>
 *    Jeff Nathan <jeffnathan (at) gmail.com>
 *
 * $Id$
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <glib.h>

#include <gtk/gtk.h>

#include <graphviz/gvc.h>
#include <graphviz/gvplugin.h>

#include <epan/packet.h>
#include <epan/report_err.h>
#include <epan/tap.h>

#include <gtk/filter_autocomplete.h>
#include <gtk/filter_dlg.h>
#include <stat_menu.h>
#include <gtk/gui_stat_menu.h>

#include <gtk/keys.h>
#include <gtk/gtkglobals.h>

#include "conv.h"

extern gvplugin_library_t gv_wireviz_gtk;

static const char *
ct_get_filter_addr(const address *addr)
{
	switch(addr->type){
		case AT_ETHER:
			return "eth.addr";
		case AT_IPv4:
			return "ip.addr";
		case AT_IPv6:
			return "ipv6.addr";
		default:
			return NULL;
	}
}

static const char *
ct_get_filter_port(int port_type_val)
{
	switch(port_type_val){
		case PT_TCP:
			return "tcp.port";
		case PT_UDP:
			return "udp.port";
		default:
			return NULL;
	}
	return NULL;
}

typedef enum {
	COL_NODE_ATTR_FILTER = 0,
	COL_NODE_ATTR_SHAPE,
	COL_NODE_ATTR_CUSTOM_TYPE,
	COL_NODE_ATTR_CUSTOM_VAL,

	COL_NODE_ATTR_ROWCOLOR,

	COL_NODE_ATTR_COUNT
} node_attrs_col_t;

typedef struct {
	dfilter_t *sfcode;
	char *filter;

	char *shape;
	char *attr;
	char *val;
} node_attr_t;

typedef struct {
	Agraph_t *g;

	Agsym_t *node_attr_filter;
	Agsym_t *node_attr_tooltip;
	Agsym_t *edge_attr_filter;
	Agsym_t *edge_attr_tooltip;

	node_attr_t *attrs;
	guint attrs_count;

} graph_data_t;

static int 
agsymset(void *obj, Agsym_t *sym, char *val)
{
	return agxset(obj, sym->index, val);
}

static void
graph_cb(gpointer key _U_, gpointer value, gpointer user_data)
{
	graph_data_t *gdata = (graph_data_t *) user_data;
	const conv_data_t *conv = value;
	char src_addr[MAX_ADDR_STR_LEN];
	char dst_addr[MAX_ADDR_STR_LEN];

	Agnode_t *a, *b;

	char *src_filter, *dest_filter;
	char *src_label, *dest_label;

	address_to_str_buf(&conv->key.src_address, src_addr, MAX_ADDR_STR_LEN);
	address_to_str_buf(&conv->key.dst_address, dst_addr, MAX_ADDR_STR_LEN);

	if (conv->port_type != PT_NONE) {
		src_label = ep_strdup_printf("%s:%u", src_addr, conv->key.src_port);
		dest_label = ep_strdup_printf("%s:%u", dst_addr, conv->key.dst_port);

		src_filter = ep_strdup_printf("%s==%s && %s == %u", ct_get_filter_addr(&conv->key.src_address), src_addr, ct_get_filter_port(conv->port_type), conv->key.src_port);
		dest_filter = ep_strdup_printf("%s==%s && %s == %u", ct_get_filter_addr(&conv->key.dst_address), dst_addr, ct_get_filter_port(conv->port_type), conv->key.dst_port);
	} else {
		src_label = src_addr;
		dest_label = dst_addr;

		src_filter = ep_strdup_printf("%s==%s", ct_get_filter_addr(&conv->key.src_address), src_addr);
		dest_filter = ep_strdup_printf("%s==%s", ct_get_filter_addr(&conv->key.dst_address), dst_addr);
	}

	a = agnode(gdata->g, ep_strdup_printf("%s_%u", src_addr, conv->key.src_port));
	agset(a, "label", src_label);
	agsymset(a, gdata->node_attr_filter, src_filter);
	agsymset(a, gdata->node_attr_tooltip, src_label);

	b = agnode(gdata->g, ep_strdup_printf("%s_%u", dst_addr, conv->key.dst_port));
	agset(b, "label", dest_label);
	agsymset(b, gdata->node_attr_filter, dest_filter);
	agsymset(b, gdata->node_attr_tooltip, dest_label);

#if 0
	if (conv->rx_frames) {
		Agedge_t *e = agedge(g, a, b);
		agsafeset(e, "label", ep_strdup_printf("%u (%llu B)", conv->rx_frames, conv->rx_bytes), "");
		agsafeset(e, "color", "darkgreen", "black");
		agsafeset(e, "fontcolor", "darkgreen", "black");
	}

	if (conv->tx_frames) {
		Agedge_t *e = agedge(g, b, a);
		agsafeset(e, "label", ep_strdup_printf("%u (%llu B)", conv->tx_frames, conv->tx_bytes), "");
		agsafeset(e, "color", "blue", "black");
		agsafeset(e, "fontcolor", "blue", "black");
	}
#endif
	{
		Agedge_t *e = agedge(gdata->g, b, a);

		agsafeset(e, "label", ep_strdup_printf("%lu/%lu", conv->rx_bytes, conv->tx_bytes), "");
		agsafeset(e, "dir", "both", "");

		agsymset(e, gdata->edge_attr_filter, ep_strdup_printf("%s && %s", src_filter, dest_filter));
		agsymset(e, gdata->edge_attr_tooltip, ep_strdup_printf(
				"%s <==> %s\n"
				"%s - %s (%.3gs)\n"
				"RX: %u (%lu B)\n"
				"TX: %u (%lu B)\n"
				"Total: %u (%lu B)", 
				src_label, dest_label,
				rel_time_to_secs_str(&conv->start_time), rel_time_to_secs_str(&conv->stop_time), 
				nstime_to_sec(&conv->stop_time) - nstime_to_sec(&conv->start_time),
				conv->rx_frames, conv->rx_bytes, conv->tx_frames, conv->tx_bytes, conv->rx_frames + conv->tx_frames, conv->rx_bytes + conv->tx_bytes));

#if 0
		conv_data_t *conv_edge;
		conv_edge = agbindrec(e, "gv_conv", sizeof(conv_data_t), FALSE);
		*conv_edge = *conv;
#endif
	}
}

static const char *layouts[] =
	{ "twopi", "circo", "dot" };

static const char *types[] =
	{ "none", "tcp", "udp", "ip", "ipv6", "eth" };

static void
graph_attr_cb(conv_data_t *conv, epan_dissect_t *edt, gpointer user_data)
{
	graph_data_t *gdata = (graph_data_t *) user_data;
	Agnode_t *a, *b;

	char src_addr[MAX_ADDR_STR_LEN];
	char dst_addr[MAX_ADDR_STR_LEN];

	guint i;

	address_to_str_buf(&conv->key.src_address, src_addr, MAX_ADDR_STR_LEN);
	address_to_str_buf(&conv->key.dst_address, dst_addr, MAX_ADDR_STR_LEN);

	a = agnode(gdata->g, ep_strdup_printf("%s_%u", src_addr, conv->key.src_port));
	b = agnode(gdata->g, ep_strdup_printf("%s_%u", dst_addr, conv->key.dst_port));

	for (i = 0; i < gdata->attrs_count; i++) {
		if (!gdata->attrs[i].sfcode || dfilter_apply_edt(gdata->attrs[i].sfcode, edt)) {
			if (gdata->attrs[i].shape) {
				agsafeset(a, "shape", gdata->attrs[i].shape, "");
				agsafeset(b, "shape", gdata->attrs[i].shape, "");
			}
			if (gdata->attrs[i].attr) {
				agsafeset(a, gdata->attrs[i].attr, gdata->attrs[i].val, "");
				agsafeset(b, gdata->attrs[i].attr, gdata->attrs[i].val, "");
			}
		}
	}
}

static gboolean
render(const char *type, const char *layout, const char *filter, node_attr_t *attrs, guint attrs_count, GtkWidget *drawingarea)
{
	static volatile gboolean rerender = FALSE; /* XXX */

	graph_data_t data;
	Agraph_t *g;
	GVC_t *gvc;
	guint i;

	GMainLoop *loop;

	if (!type || !layout || !drawingarea) {
		g_warning("render: %p %p %p\n", type, layout, drawingarea);
		return FALSE;
	}

	if ((loop = g_object_get_data(G_OBJECT(drawingarea), "loop"))) {
		if (g_main_loop_is_running(loop)) {
			g_main_loop_quit(loop);
			rerender = TRUE;
			return FALSE;
		}
	}

	gvc = gvContext();
	gvAddLibrary(gvc, &gv_wireviz_gtk);
	
	data.g = g = agopen("g", AGDIGRAPH);
	data.attrs = attrs;
	data.attrs_count = attrs_count;

	/* default values */
	agraphattr(g, "ratio", "auto");
	agraphattr(g, "ranksep", "3");

	/* This one SiGSEGVs AFAIR on null-derefence.
	extern attrsym_t *G_activefillcolor;
	extern attrsym_t *G_activepencolor;
	agsymset(g, G_activefillcolor, "green");
	agsymset(g, G_activepencolor, "green");
	*/

	data.node_attr_filter = agnodeattr(g, "filter", "");
	data.node_attr_tooltip = agnodeattr(g, "tooltip", "");
	data.edge_attr_filter = agedgeattr(g, "filter", "");
	data.edge_attr_tooltip = agedgeattr(g, "tooltip", "");

	/* build connection graph */
	if (data.attrs) {
		for (i = 0; i < attrs_count; i++)
			register_tap_listener("frame", &data.attrs[i], data.attrs[i].filter, 0, NULL, NULL, NULL);
		conv_list(type, filter, graph_cb, graph_attr_cb, &data);

		for (i = 0; i < attrs_count; i++) 
			remove_tap_listener(&data.attrs[i]);
	} else
		conv_list(type, filter, graph_cb, NULL, &data);

	/* layout */
	gvLayout(gvc, g, layout);

	/* render */
	gvRender(gvc, g, "gviz", drawingarea);
	/* XXX, if (rerender) and attrs/type hasn't change it's enough to do gvRender once again */

	/* free memory */
	gvFreeLayout(gvc, g);
	agclose(g);
	gvFreeContext(gvc);

	if (rerender) {
		rerender = FALSE;
		return TRUE;
	}
	return FALSE;
}

typedef struct {
	GtkTreeModel *model;
	GtkWidget *filter;
	GtkWidget *lay, *typ;
	GtkWidget *drawingarea;

} render_cb_t;

static render_cb_t render_data; // XXX

static void
render_cb(GtkWidget *widget _U_, gpointer user_data)
{
	render_cb_t *data = (render_cb_t *) user_data;
	gboolean ret = TRUE;
	
	do {
		node_attr_t *attrs = NULL;
		GtkTreeIter iter;
		int count = 0;
		int i;

		if (gtk_tree_model_get_iter_first(data->model, &iter)) {
			count = 1;
			while (gtk_tree_model_iter_next(data->model, &iter))
				count++;

			attrs = g_malloc0(sizeof(node_attr_t) * count);

			gtk_tree_model_get_iter_first(data->model, &iter);
			for (i = 0; i < count; i++) {
				gtk_tree_model_get(data->model, &iter, 
						COL_NODE_ATTR_FILTER, &attrs[i].filter,
						COL_NODE_ATTR_SHAPE, &attrs[i].shape,
						COL_NODE_ATTR_CUSTOM_TYPE, &attrs[i].attr, 
						COL_NODE_ATTR_CUSTOM_VAL, &attrs[i].val,
						-1);

				if (attrs[i].filter && !dfilter_compile(attrs[i].filter, &attrs[i].sfcode)) {
					report_failure("Compiling filter: %s (index: %d) failed!\n", attrs[i].filter, i);
					ret = FALSE;
					break;
				}
				gtk_tree_model_iter_next(data->model, &iter);
			}
		}

		if (ret)
			ret = render(gtk_combo_box_get_active_text(GTK_COMBO_BOX(data->typ)), gtk_combo_box_get_active_text(GTK_COMBO_BOX(data->lay)), gtk_entry_get_text(GTK_ENTRY(data->filter)), attrs, count, data->drawingarea);

		for (i = 0; i < count; i++) {
			g_free(attrs[i].filter);
			g_free(attrs[i].attr);
			g_free(attrs[i].val);
			g_free(attrs[i].shape);

			if (attrs[i].sfcode)
				dfilter_free(attrs[i].sfcode);
		}
		g_free(attrs);

	} while (ret);
}

static void
attr_edited_common(GtkTreeModel *model, int column, gchar *path_string, gchar *new_text)
{
	GtkTreeIter iter;

	gtk_tree_model_get_iter_from_string(model, &iter, path_string);
	gtk_list_store_set(GTK_LIST_STORE(model), &iter, column, new_text, -1);
}

static void
attr_filter_edited_cb(GtkCellRendererText *cell _U_, gchar *path_string, gchar *filter, gpointer user_data)
{
	static const char deprecated_color[] = "#ffffb0";
	static const char valid_color[] = "#b0ffb0";
	static const char invalid_color[] = "#ffb0b0";

	GtkTreeModel *model = GTK_TREE_MODEL(user_data);
	GtkTreeIter iter;
	const char *color;

	gtk_tree_model_get_iter_from_string(model, &iter, path_string);

	/* row color */
	if (filter && filter[0]) {
		dfilter_t *dfp;

		if (dfilter_compile(filter, &dfp)) {
			if (dfilter_deprecated_tokens(dfp))
				color = deprecated_color;
			else
				color = valid_color;
			dfilter_free(dfp);
		} else
			color = invalid_color;
	} else
		color = NULL;

	gtk_list_store_set(GTK_LIST_STORE(model), &iter, COL_NODE_ATTR_FILTER, filter, COL_NODE_ATTR_ROWCOLOR, color, -1);
}

static void
attr_filter_edited_start_cb(GtkCellRenderer *cell _U_, GtkCellEditable *editable, const gchar *path _U_, gpointer data _U_)
{
	if (GTK_IS_ENTRY(editable)) {
		GtkEntry *entry = GTK_ENTRY(editable);

		filter_te_syntax_check_cb(GTK_WIDGET(entry), NULL);
		g_signal_connect(entry, "changed", G_CALLBACK(filter_te_syntax_check_cb), NULL);
		g_signal_connect(entry, "key-press-event", G_CALLBACK(filter_string_te_key_pressed_cb), NULL);
	}
}

static void
attr_shape_edited_cb(GtkCellRendererText *cell _U_, gchar *path_string, gchar *new_text, gpointer user_data)
{
	attr_edited_common(GTK_TREE_MODEL(user_data), COL_NODE_ATTR_SHAPE, path_string, new_text);
}

static void
attr_attr_edited_cb(GtkCellRendererText *cell _U_, gchar *path_string, gchar *new_text, gpointer user_data)
{
	attr_edited_common(GTK_TREE_MODEL(user_data), COL_NODE_ATTR_CUSTOM_TYPE, path_string, new_text);
}

static void
attr_value_edited_cb(GtkCellRendererText *cell _U_, gchar *path_string, gchar *new_text, gpointer user_data)
{
	attr_edited_common(GTK_TREE_MODEL(user_data), COL_NODE_ATTR_CUSTOM_VAL, path_string, new_text);
}

static void
wireviz_cb(void)
{
	/* http://www.graphviz.org/doc/info/attrs.html | grep 'N' */
	static const char *attrs[] = { 
		NULL,
		"area", "color", "colorscheme", "distortion", "fillcolor", "fixedsize", "fontcolor", "fontname", "fontsize", "group", "height", "image", "imagescale" 
		"labelloc", "margin", "nojustify", "ordering", "orientation", "penwidth", "peripheries", "pin", "pos", "rects", "regular", "root", "samplepoints", "shape", "shapefile", 
		"sides", "skew", "sortv", "style", "vertices", "width", "z"
		/* ps-only: layer, showboxes */
		/* "comment", "label", "URL", "href", "id", "tooltip", "target" */
	};

	/* http://www.graphviz.org/doc/info/shapes.html */
	static const char *shapes[] = {
		NULL,
		"box", "polygon", "ellipse", "oval", "circle", "point", "egg", "triangle",
		"plaintext", "diamond", "trapezium", "parallelogram",
		"house", "pentagon", "hexagon", "septagon",
		"octagon", "doublecircle", "doubleoctagon", 
		"tripleoctagon", "invtriangle", "invtrapezium", "invhouse",
		"Mdiamond", "Msquare", "Mcircle", "rect", "rectangle",
		"square", "none", "note", "tab", "folder", "box3d", "component"
	};

	GtkWidget *window, *vbox, *hbox, *tree_expand, *tree, *drawingarea;
	GtkWidget *type, *layout, *filter, *render;
	GtkWidget *profile_bar;
	GtkListStore *model, *attr_model, *shapes_model;
	GtkCellRenderer *renderer;
	guint i;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), "WireViz");

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(window), vbox);

	hbox = gtk_hbox_new(FALSE, 3);

	type = gtk_combo_box_new_text();
	for (i = 0; i < array_length(types); i++)
		gtk_combo_box_append_text(GTK_COMBO_BOX(type), types[i]);
	gtk_combo_box_set_active(GTK_COMBO_BOX(type), 1 /* tcp */);
	gtk_combo_box_set_wrap_width(GTK_COMBO_BOX(type),  array_length(types));
	gtk_box_pack_start(GTK_BOX(hbox), type, FALSE, FALSE, 0);
	gtk_widget_show(type);

	layout = gtk_combo_box_new_text();
	for (i = 0; i < array_length(layouts); i++)
		gtk_combo_box_append_text(GTK_COMBO_BOX(layout), layouts[i]);
	gtk_combo_box_set_active(GTK_COMBO_BOX(layout), 0);
	gtk_combo_box_set_wrap_width(GTK_COMBO_BOX(layout),  array_length(layouts));
	gtk_box_pack_start(GTK_BOX(hbox), layout, FALSE, FALSE, 0);
	gtk_widget_show(layout);

	filter = gtk_entry_new();
	g_signal_connect(filter, "changed", G_CALLBACK(filter_te_syntax_check_cb), NULL);
	g_signal_connect(filter, "key-press-event", G_CALLBACK(filter_string_te_key_pressed_cb), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), filter, TRUE, TRUE, 0);
	gtk_widget_show(filter);

	render = gtk_button_new_from_stock(GTK_STOCK_EXECUTE);
	gtk_button_set_label(GTK_BUTTON(render), "Render graph");
	gtk_box_pack_start(GTK_BOX(hbox), render, FALSE, FALSE, 0);
	gtk_widget_show(render);

	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	attr_model = gtk_list_store_new(1, G_TYPE_STRING);
	for (i = 0; i < array_length(attrs); i++)
		gtk_list_store_insert_with_values(attr_model, NULL, i, 0, attrs[i], -1);

	shapes_model = gtk_list_store_new(1, G_TYPE_STRING);
	for (i = 0; i < array_length(shapes); i++)
		gtk_list_store_insert_with_values(shapes_model, NULL, i, 0, shapes[i], -1);

	model = gtk_list_store_new(COL_NODE_ATTR_COUNT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
	for (i = 0; i < 6; i++) { /* XXX, 5 */
		GtkTreeIter iter;
		gtk_list_store_append(model, &iter);
	}

	tree = gtk_tree_view_new();
	gtk_tree_view_set_model(GTK_TREE_VIEW(tree), GTK_TREE_MODEL(model));

	renderer = gtk_cell_renderer_text_new();
	g_object_set(renderer, "editable", TRUE, NULL);
	g_signal_connect(renderer, "edited", G_CALLBACK(attr_filter_edited_cb), model);
	g_signal_connect(renderer, "editing-started", G_CALLBACK(attr_filter_edited_start_cb), NULL);
	/* copy main display filter */
	{
			GtkWidget *filter_cm = g_object_get_data(G_OBJECT(top_level), E_DFILTER_CM_KEY);
			GtkWidget *filter_te = gtk_bin_get_child(GTK_BIN(filter_cm));

			gtk_entry_set_text(GTK_ENTRY(filter), gtk_entry_get_text(GTK_ENTRY(filter_te)));
	}
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), 0, "Filter", renderer, "text", COL_NODE_ATTR_FILTER, "background", COL_NODE_ATTR_ROWCOLOR, NULL);

	renderer = gtk_cell_renderer_combo_new();
	g_object_set(renderer, "editable", TRUE, "model", shapes_model, "text-column", 0, "has-entry", FALSE, NULL);
	g_signal_connect(renderer, "edited", (GCallback) attr_shape_edited_cb, model);
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), 1, "Shape", renderer, "text", COL_NODE_ATTR_SHAPE, "background", COL_NODE_ATTR_ROWCOLOR, NULL);

	renderer = gtk_cell_renderer_combo_new();
	g_object_set(renderer, "editable", TRUE, "model", attr_model, "text-column", 0, "has-entry", FALSE, NULL);
	g_signal_connect(renderer, "edited", (GCallback) attr_attr_edited_cb, model);
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), 2, "Custom Attribute", renderer, "text", COL_NODE_ATTR_CUSTOM_TYPE, "background", COL_NODE_ATTR_ROWCOLOR, NULL);

	/* renderer = gtk_cell_renderer_combo_new(); */
	renderer = gtk_cell_renderer_text_new();
	g_object_set(renderer, "editable", TRUE, NULL);
	g_signal_connect(renderer, "edited", (GCallback) attr_value_edited_cb, model);
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), 3, "Attribute Value", renderer, "text", COL_NODE_ATTR_CUSTOM_VAL, "background", COL_NODE_ATTR_ROWCOLOR, NULL);

	gtk_widget_show(tree);
	/* gtk_tree_view_set_grid_lines(GTK_TREE_VIEW(tree), GTK_TREE_VIEW_GRID_LINES_BOTH); */
	gtk_tree_view_columns_autosize(GTK_TREE_VIEW(tree));

	tree_expand = gtk_expander_new("Extra node attributes");
	g_object_set(tree_expand, "border-width", 2, NULL);
	gtk_container_add(GTK_CONTAINER(tree_expand), tree);
	gtk_widget_show(tree_expand);

	gtk_box_pack_start(GTK_BOX(vbox), tree_expand, FALSE, FALSE, 0);

	drawingarea = gtk_drawing_area_new();
	gtk_widget_show(drawingarea);
	gtk_box_pack_start(GTK_BOX(vbox), drawingarea, TRUE, TRUE, 0);
	gtk_widget_set_size_request(drawingarea, 300, 300);
	gtk_widget_set_events(drawingarea, GDK_EXPOSURE_MASK | GDK_POINTER_MOTION_MASK | GDK_BUTTON_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK);

	profile_bar = gtk_statusbar_new();
	gtk_box_pack_start(GTK_BOX(vbox), profile_bar, FALSE, FALSE, 0);
	gtk_widget_show(profile_bar);

	gtk_widget_show(window);

	render_data.filter = filter;
	render_data.lay = layout;
	render_data.typ = type;
	render_data.model = GTK_TREE_MODEL(model);
	render_data.drawingarea = drawingarea;

	g_signal_connect(render, "clicked", G_CALLBACK(render_cb), &render_data);
}

extern
void
proto_reg_handoff_wireviz(void)
{
}


extern
void
proto_register_wireviz(void)
{
	register_stat_menu_item("WireViz", REGISTER_STAT_GROUP_GENERIC, wireviz_cb, NULL, NULL, NULL);
}

